from django.urls import include, path

# Default URL patterns - you can break this out in the individual modules if preferred.

urlpatterns = [
    path("digid/", include("digid_eherkenning.digid_urls")),
    path("eherkenning/", include("digid_eherkenning.eherkenning_urls")),
    path("metadata/", include("digid_eherkenning.metadata_urls")),
]



from io import StringIO
from unittest.mock import patch

from django.core.management import CommandError, call_command
from django.test import TestCase

import pytest

from digid_eherkenning.models import DigidConfiguration, EherkenningConfiguration

from .conftest import DIGID_TEST_METADATA_FILE_SLO_POST, EHERKENNING_TEST_METADATA_FILE


@pytest.mark.usefixtures("digid_config_defaults", "temp_private_root")
class UpdateStoredMetadataTests(TestCase):
    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_command_triggers_xml_fetching_when_digid(self, get_metadata):
        output = StringIO()
        config = DigidConfiguration.get_solo()

        with DIGID_TEST_METADATA_FILE_SLO_POST.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_metadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://was-preprod1.digid.nl/saml/idp/metadata"
            )
            config.save()

            self.assertEqual(get_metadata.call_count, 1)

            get_metadata.reset_mock()

            call_command("update_stored_metadata", "digid", stdout=output)

        self.assertEqual(get_metadata.call_count, 1)
        self.assertEqual(output.getvalue(), "Update was successful\n")

    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_command_triggers_xml_fetching_when_eherkenning(self, get_metadata):
        output = StringIO()
        config = EherkenningConfiguration.get_solo()

        with EHERKENNING_TEST_METADATA_FILE.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_metadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://eh01.staging.iwelcome.nl/broker/sso/1.13"
            )
            config.save()

            self.assertEqual(get_metadata.call_count, 1)

            get_metadata.reset_mock()

            call_command("update_stored_metadata", "eherkenning", stdout=output)

        self.assertEqual(get_metadata.call_count, 1)
        self.assertEqual(output.getvalue(), "Update was successful\n")

    def test_command_fails_when_no_argument_provided(self):
        try:
            call_command("update_stored_metadata")
        except CommandError as e:
            error_message = str(e)

        self.assertEqual(
            error_message,
            "Error: the following arguments are required: config_model",
        )

    def test_command_fails_when_no_metadata_file_source(self):
        output = StringIO()
        call_command("update_stored_metadata", "digid", stdout=output)

        self.assertEqual(
            output.getvalue(),
            "Update failed, no metadata file source found\n",
        )



from pathlib import Path
from typing import Optional, TypedDict, Union


class ServiceConfig(TypedDict):
    service_uuid: str
    service_name: str
    attribute_consuming_service_index: str
    service_instance_uuid: str
    service_description: str
    service_description_url: str
    service_url: str
    loa: str
    privacy_policy_url: str
    herkenningsmakelaars_id: str
    requested_attributes: str
    service_restrictions_allowed: str
    entity_concerned_types_allowed: list[dict]
    language: str
    classifiers: Optional[list[str]]


class EHerkenningConfig(TypedDict):
    base_url: str
    acs_path: str
    entity_id: str
    metadata_file: str
    cert_file: Path
    key_file: Path
    service_entity_id: str
    oin: str
    services: list[ServiceConfig]
    want_assertions_encrypted: str
    want_assertions_signed: str
    key_passphrase: str
    signature_algorithm: str
    digest_algorithm: str
    technical_contact_person_telephone: Optional[str]
    technical_contact_person_email: Optional[str]
    organization: str
    organization_name: str
    artifact_resolve_content_type: str


class ServiceProviderSAMLConfig(TypedDict):
    entityId: str
    assertionConsumerService: dict
    singleLogoutService: dict
    attributeConsumingServices: list[dict]
    NameIDFormat: str
    x509cert: str
    privateKey: str
    privateKeyPassphrase: Optional[str]


class IdentityProviderSAMLConfig(TypedDict):
    entityId: str
    singleSignOnService: dict
    singleLogoutService: dict
    x509cert: str


class SecuritySAMLConfig(TypedDict):
    nameIdEncrypted: bool
    authnRequestsSigned: bool
    logoutRequestSigned: bool
    logoutResponseSigned: bool
    signMetadata: bool
    wantMessagesSigned: bool
    wantAssertionsSigned: bool
    wantAssertionsEncrypted: bool
    wantNameId: bool
    wantNameIdEncrypted: bool
    wantAttributeStatement: bool
    requestedAuthnContext: Union[bool, list[str]]
    requestedAuthnContextComparison: str
    failOnAuthnContextMismatch: bool
    metadataValidUntil: Optional[str]
    metadataCacheDuration: Optional[str]
    allowSingleLabelDomains: bool
    signatureAlgorithm: str
    digestAlgorithm: str
    allowRepeatAttributeName: bool
    rejectDeprecatedAlgorithm: bool
    disableSignatureWrappingProtection: bool


class EHerkenningSAMLConfig(TypedDict):
    strict: bool
    debug: bool
    sp: ServiceProviderSAMLConfig
    idp: IdentityProviderSAMLConfig
    security: SecuritySAMLConfig
    contactPerson: dict
    organization: dict



import logging
import re

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from digid_eherkenning.backends import BaseBackend, BSNBackendMixin

logger = logging.getLogger(__name__)

UserModel = get_user_model()


class DigiDBackend(BSNBackendMixin, BaseBackend):
    service_name = "DigiD_Mock"
    error_messages = dict(
        BaseBackend.error_messages,
        **{
            "digid_no_bsn": _("Login failed due to no BSN being returned by DigiD."),
            "digid_len_bsn": _("Login failed due to no BSN having more then 9 digits."),
            "digid_num_bsn": _("Login failed due to no BSN not being numerical."),
        }
    )

    def authenticate(self, request, bsn=None):
        if bsn is None:
            return

        if bsn == "":
            self.log_error(request, self.error_messages["digid_no_bsn"])
            return

        elif not re.match(r"^[0-9]+$", bsn):
            self.log_error(request, self.error_messages["digid_num_bsn"])
            return

        elif len(bsn) > 9:
            self.log_error(request, self.error_messages["digid_len_bsn"])
            return

        user = self.get_or_create_user_from_bsn(request, bsn)

        # DigiD requires a session of max 15 minutes. See DigiDCheck 2.2 T14 -- Sessieduur
        # request.session.set_expiry(15 * 60)

        return user



from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

# See `OINType` in eherkenning-dc.xml XSD
oin_validator = RegexValidator(
    regex=r"[0-9]{20}",
    message=_("A valid OIN consists of 20 digits."),
)


validate_digits = RegexValidator(
    regex="^[0-9]+$", message=_("Expected a numerical value.")
)


class Proef11ValidatorBase:
    value_size = NotImplemented
    error_messages = {
        "too_short": NotImplemented,
        "wrong": NotImplemented,
    }

    def __call__(self, value):
        """
        Validates that a string value is a valid 11-proef number (BSN, RSIN etc) by applying the
        '11-proef' checking.
        :param value: String object representing a presumably good 11-proef number.
        """
        # Initial sanity checks.
        validate_digits(value)
        if len(value) != self.value_size:
            raise ValidationError(
                self.error_messages["too_short"],
                params={"size": self.value_size},
                code="invalid",
            )

        # 11-proef check.
        total = 0
        for multiplier, char in enumerate(reversed(value), start=1):
            if multiplier == 1:
                total += -multiplier * int(char)
            else:
                total += multiplier * int(char)

        if total % 11 != 0:
            raise ValidationError(self.error_messages["wrong"])


@deconstructible
class BSNValidator(Proef11ValidatorBase):
    """
    Validate a BSN value by applying the "11-proef".
    """

    value_size = 9
    error_messages = {
        "too_short": _("BSN should have %(size)i characters."),
        "wrong": _("Invalid BSN."),
    }



from urllib.parse import urlencode

from django.contrib.auth import get_user_model
from django.test import TestCase, modify_settings, override_settings
from django.urls import reverse, reverse_lazy

from furl import furl


class DigidMockTestCase(TestCase):
    def assertNoDigidURLS(self, response):
        # verify no links to DigiD remain in template
        self.assertNotContains(response, "://digid.nl")
        self.assertNotContains(response, "://www.digid.nl")


OVERRIDE_SETTINGS = dict(
    DIGID_MOCK_APP_TITLE="FooBarBazz-MockApp",
    DIGID_MOCK_RETURN_URL=reverse_lazy(
        "test-success"
    ),  # url to redirect to after success
    DIGID_MOCK_CANCEL_URL=reverse_lazy(
        "test-index"
    ),  # url to navigate to when users clicks 'cancel/annuleren'
    ROOT_URLCONF="tests.project.mock_urls",
)

MODIFY_SETTINGS = dict(
    AUTHENTICATION_BACKENDS={
        "append": [
            "digid_eherkenning.mock.backends.DigiDBackend",
        ],
        "remove": [
            "digid_eherkenning.backends.DigiDBackend",
        ],
    }
)


@override_settings(**OVERRIDE_SETTINGS)
@modify_settings(**MODIFY_SETTINGS)
class TestAppIndexTests(TestCase):
    def test_get_returns_valid_response(self):
        url = reverse("test-index")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, reverse("digid:login"))


@override_settings(**OVERRIDE_SETTINGS)
@modify_settings(**MODIFY_SETTINGS)
class LoginViewTests(DigidMockTestCase):
    def test_get_returns_http400_on_missing_params(self):
        url = reverse("digid-mock:login")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 400)

    def test_get_returns_valid_response(self):
        url = reverse("digid-mock:login")
        data = {
            "acs": reverse("digid:acs"),
            "next": reverse("test-success"),
            "cancel": reverse("test-index"),
        }
        response = self.client.get(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "FooBarBazz-MockApp")
        self.assertContains(response, reverse("digid-mock:password"))
        self.assertNoDigidURLS(response)


@override_settings(**OVERRIDE_SETTINGS)
@modify_settings(**MODIFY_SETTINGS)
class PasswordLoginViewTests(DigidMockTestCase):
    def test_get_returns_http400_on_missing_params(self):
        url = reverse("digid-mock:password")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 400)

    def test_get_returns_valid_response(self):
        url = reverse("digid-mock:password")
        data = {
            "acs": reverse("digid:acs"),
            "next": reverse("test-success"),
            "cancel": reverse("test-index"),
        }
        response = self.client.get(url, data=data)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "FooBarBazz-MockApp")
        self.assertContains(response, reverse("digid-mock:login"))
        self.assertNoDigidURLS(response)

    def test_post_redirects_and_authenticates(self):
        url = reverse("digid-mock:password")
        params = {
            "acs": reverse("digid:acs"),
            "next": reverse("test-success"),
            "cancel": reverse("test-index"),
        }
        url = f"{url}?{urlencode(params)}"

        data = {
            "auth_name": "296648875",
            "auth_pass": "bar",
        }
        # post our password to the IDP
        response = self.client.post(url, data, follow=False)

        # it will redirect to our ACS
        self.assertEqual(response.status_code, 302)
        self.assertIn(reverse("digid:acs"), response["Location"])

        # follow the ACS redirect and get/create the user
        response = self.client.get(response["Location"], follow=False)

        User = get_user_model()
        user = User.digid_objects.get(bsn="296648875")

        # follow redirect to 'next'
        self.assertRedirects(response, reverse("test-success"))

        response = self.client.get(response["Location"], follow=False)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Je bent ingelogged als gebruiker")
        self.assertContains(response, "<code>{}</code>".format(str(user)))
        self.assertContains(response, "<code>296648875</code>")

    def test_post_redirect_retains_acs_querystring_params(self):
        url = reverse("digid-mock:password")
        params = {
            "acs": f"{reverse('digid:acs')}?foo=bar",
            "next": reverse("test-success"),
            "cancel": reverse("test-index"),
        }
        url = f"{url}?{urlencode(params)}"

        data = {
            "auth_name": "296648875",
            "auth_pass": "bar",
        }
        # post our password to the IDP
        response = self.client.post(url, data, follow=False)

        # it will redirect to our ACS
        expected_redirect = furl(reverse("digid:acs")).set(
            {
                "foo": "bar",
                "bsn": "296648875",
                "next": reverse("test-success"),
            }
        )
        self.assertRedirects(
            response, str(expected_redirect), fetch_redirect_response=False
        )


@override_settings(**OVERRIDE_SETTINGS)
@modify_settings(**MODIFY_SETTINGS)
class LogoutViewTests(TestCase):
    def test_logout(self):
        User = get_user_model()
        user = User.objects.create_user(username="testuser", password="test")
        self.client.force_login(user)

        url = reverse("digid:logout")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 302)
        self.assertFalse("_auth_user_id" in self.client.session)



# Generated by Django 3.2.15 on 2022-10-27 15:08

import uuid

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import privates.fields
import privates.storages

import digid_eherkenning.models.digid


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("simple_certmanager", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="EherkenningConfiguration",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "idp_metadata_file",
                    privates.fields.PrivateMediaFileField(
                        help_text="The metadata file of the identity provider.",
                        storage=privates.storages.PrivateMediaFileSystemStorage(),
                        upload_to="",
                        verbose_name="identity provider metadata",
                    ),
                ),
                (
                    "idp_service_entity_id",
                    models.CharField(
                        help_text="Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note that this must match the 'entityID' attribute on the 'md:EntityDescriptor' node found in the Identity Provider's metadata.",
                        max_length=255,
                        verbose_name="identity provider service entity ID",
                    ),
                ),
                (
                    "want_assertions_signed",
                    models.BooleanField(
                        default=True,
                        help_text="If True, the XML assertions need to be signed, otherwise the whole response needs to be signed.",
                        max_length=100,
                        verbose_name="want assertions signed",
                    ),
                ),
                (
                    "want_assertions_encrypted",
                    models.BooleanField(
                        default=False,
                        help_text="If True the XML assertions need to be encrypted.",
                        max_length=100,
                        verbose_name="want assertions encrypted",
                    ),
                ),
                (
                    "key_passphrase",
                    models.CharField(
                        blank=True,
                        help_text="Passphrase for the private key used by the SOAP client.",
                        max_length=100,
                        verbose_name="key passphrase",
                    ),
                ),
                (
                    "signature_algorithm",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("http://www.w3.org/2000/09/xmldsig#dsa-sha1", "DSA_SHA1"),
                            ("http://www.w3.org/2000/09/xmldsig#rsa-sha1", "RSA_SHA1"),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
                                "RSA_SHA256",
                            ),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha384",
                                "RSA_SHA384",
                            ),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha512",
                                "RSA_SHA512",
                            ),
                        ],
                        default="http://www.w3.org/2000/09/xmldsig#rsa-sha1",
                        help_text="Signature algorithm. Note that DSA_SHA1 and RSA_SHA1 are deprecated, but RSA_SHA1 is still the default value in the SAMLv2 standard. Warning: there are known issues with single-logout functionality if using anything other than SHA1 due to some hardcoded algorithm.",
                        max_length=100,
                        verbose_name="signature algorithm",
                    ),
                ),
                (
                    "digest_algorithm",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("http://www.w3.org/2000/09/xmldsig#sha1", "SHA1"),
                            ("http://www.w3.org/2001/04/xmlenc#sha256", "SHA256"),
                            ("http://www.w3.org/2001/04/xmldsig-more#sha384", "SHA384"),
                            ("http://www.w3.org/2001/04/xmlenc#sha512", "SHA512"),
                        ],
                        default="http://www.w3.org/2000/09/xmldsig#sha1",
                        help_text="Digest algorithm. Note that SHA1 is deprecated, but still the default value in the SAMLv2 standard. Warning: there are known issues with single-logout functionality if using anything other than SHA1 due to some hardcoded algorithm.",
                        max_length=100,
                        verbose_name="digest algorithm",
                    ),
                ),
                (
                    "entity_id",
                    models.CharField(
                        help_text="Service provider entity ID.",
                        max_length=100,
                        verbose_name="entity ID",
                    ),
                ),
                (
                    "base_url",
                    models.URLField(
                        help_text="Base URL of the application.",
                        max_length=100,
                        verbose_name="base URL",
                    ),
                ),
                (
                    "service_name",
                    models.CharField(
                        help_text="Name of the service you are providing.",
                        max_length=100,
                        verbose_name="service name",
                    ),
                ),
                (
                    "service_description",
                    models.CharField(
                        help_text="A description of the service you are providing.",
                        max_length=100,
                        verbose_name="service description",
                    ),
                ),
                (
                    "technical_contact_person_telephone",
                    models.CharField(
                        blank=True,
                        help_text="Telephone number of the technical person responsible for this DigiD/eHerkenning/eIDAS setup. For it to show up in the metata, you should also specify the email address.",
                        max_length=100,
                        verbose_name="technical contact: phone number",
                    ),
                ),
                (
                    "technical_contact_person_email",
                    models.CharField(
                        blank=True,
                        help_text="Email address of the technical person responsible for this DigiD/eHerkenning/eIDAS setup. For it to show up in the metadata, you should also specify the phone number.",
                        max_length=100,
                        verbose_name="technical contact: email",
                    ),
                ),
                (
                    "organization_url",
                    models.URLField(
                        blank=True,
                        help_text="URL of the organization providing the service for which DigiD/eHerkenning/eIDAS login is configured. For it to show up in the metadata, you should also specify the organization name.",
                        max_length=255,
                        verbose_name="organization URL",
                    ),
                ),
                (
                    "organization_name",
                    models.CharField(
                        blank=True,
                        help_text="URL of the organization providing the service for which DigiD/eHerkenning/eIDAS login is configured. For it to show up in the metadata, you should also specify the organization URL.",
                        max_length=100,
                        verbose_name="organization name",
                    ),
                ),
                (
                    "loa",
                    models.CharField(
                        blank=True,
                        choices=[
                            (
                                "urn:etoegang:core:assurance-class:loa1",
                                "Non existent (1)",
                            ),
                            ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                            ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                            (
                                "urn:etoegang:core:assurance-class:loa3",
                                "Substantial (3)",
                            ),
                            ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                        ],
                        default="urn:etoegang:core:assurance-class:loa3",
                        help_text="Level of Assurance (LoA) to use for all the services.",
                        max_length=100,
                        verbose_name="LoA",
                    ),
                ),
                (
                    "eh_attribute_consuming_service_index",
                    models.CharField(
                        blank=True,
                        default="9052",
                        help_text="Attribute consuming service index for the eHerkenning service",
                        max_length=100,
                        verbose_name="eHerkenning attribute consuming service index",
                    ),
                ),
                (
                    "eh_requested_attributes",
                    models.JSONField(
                        default=list,
                        help_text="A list of additional requested attributes. A single requested attribute can be a string (the name of the attribute) or an object with keys 'name' and 'required', where 'name' is a string and 'required' a boolean'.",
                        verbose_name="requested attributes",
                    ),
                ),
                (
                    "eh_service_uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="UUID of the eHerkenning service. Once entered into catalogues, changing the value is a manual process.",
                        verbose_name="eHerkenning service UUID",
                    ),
                ),
                (
                    "eh_service_instance_uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="UUID of the eHerkenning service instance. Once entered into catalogues, changing the value is a manual process.",
                        verbose_name="eHerkenning service UUID",
                    ),
                ),
                (
                    "eidas_attribute_consuming_service_index",
                    models.CharField(
                        blank=True,
                        default="9053",
                        help_text="Attribute consuming service index for the eIDAS service",
                        max_length=100,
                        verbose_name="eIDAS attribute consuming service index",
                    ),
                ),
                (
                    "eidas_requested_attributes",
                    models.JSONField(
                        default=list,
                        help_text="A list of additional requested attributes. A single requested attribute can be a string (the name of the attribute) or an object with keys 'name' and 'required', where 'name' is a string and 'required' a boolean'.",
                        verbose_name="requested attributes",
                    ),
                ),
                (
                    "eidas_service_uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="UUID of the eIDAS service. Once entered into catalogues, changing the value is a manual process.",
                        verbose_name="eIDAS service UUID",
                    ),
                ),
                (
                    "eidas_service_instance_uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="UUID of the eHerkenning service instance. Once entered into catalogues, changing the value is a manual process.",
                        verbose_name="eIDAS service UUID",
                    ),
                ),
                (
                    "oin",
                    models.CharField(
                        help_text="The OIN of the company providing the service.",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                message="A valid OIN consists of 20 digits.",
                                regex="[0-9]{20}",
                            )
                        ],
                        verbose_name="OIN",
                    ),
                ),
                (
                    "no_eidas",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="If True, then the service catalogue will contain only the eHerkenning service.",
                        verbose_name="no eIDAS",
                    ),
                ),
                (
                    "privacy_policy",
                    models.URLField(
                        help_text="The URL where the privacy policy from the organization providing the service can be found.",
                        max_length=255,
                        verbose_name="privacy policy",
                    ),
                ),
                (
                    "makelaar_id",
                    models.CharField(
                        help_text="OIN of the broker used to set up eHerkenning/eIDAS.",
                        max_length=100,
                        validators=[
                            django.core.validators.RegexValidator(
                                message="A valid OIN consists of 20 digits.",
                                regex="[0-9]{20}",
                            )
                        ],
                        verbose_name="broker ID",
                    ),
                ),
                (
                    "artifact_resolve_content_type",
                    models.CharField(
                        choices=[
                            ("application/soap+xml", "application/soap+xml"),
                            ("text/xml", "text/xml"),
                        ],
                        default="application/soap+xml",
                        help_text="'application/soap+xml' is considered legacy and modern brokers typically expect 'text/xml'.",
                        max_length=100,
                        verbose_name="resolve artifact binding content type",
                    ),
                ),
                (
                    "service_language",
                    models.CharField(
                        default="nl",
                        help_text="Metadata for eHerkenning/eidas will contain this language key",
                        max_length=2,
                        verbose_name="service language",
                    ),
                ),
                (
                    "certificate",
                    models.ForeignKey(
                        help_text="The private key and public certificate pair to use during the authentication flow.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="simple_certmanager.certificate",
                        verbose_name="key pair",
                    ),
                ),
            ],
            options={
                "verbose_name": "Eherkenning/eIDAS metadata configuration",
            },
        ),
        migrations.CreateModel(
            name="DigidConfiguration",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "idp_metadata_file",
                    privates.fields.PrivateMediaFileField(
                        help_text="The metadata file of the identity provider.",
                        storage=privates.storages.PrivateMediaFileSystemStorage(),
                        upload_to="",
                        verbose_name="identity provider metadata",
                    ),
                ),
                (
                    "idp_service_entity_id",
                    models.CharField(
                        help_text="Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note that this must match the 'entityID' attribute on the 'md:EntityDescriptor' node found in the Identity Provider's metadata.",
                        max_length=255,
                        verbose_name="identity provider service entity ID",
                    ),
                ),
                (
                    "want_assertions_signed",
                    models.BooleanField(
                        default=True,
                        help_text="If True, the XML assertions need to be signed, otherwise the whole response needs to be signed.",
                        max_length=100,
                        verbose_name="want assertions signed",
                    ),
                ),
                (
                    "want_assertions_encrypted",
                    models.BooleanField(
                        default=False,
                        help_text="If True the XML assertions need to be encrypted.",
                        max_length=100,
                        verbose_name="want assertions encrypted",
                    ),
                ),
                (
                    "key_passphrase",
                    models.CharField(
                        blank=True,
                        help_text="Passphrase for the private key used by the SOAP client.",
                        max_length=100,
                        verbose_name="key passphrase",
                    ),
                ),
                (
                    "signature_algorithm",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("http://www.w3.org/2000/09/xmldsig#dsa-sha1", "DSA_SHA1"),
                            ("http://www.w3.org/2000/09/xmldsig#rsa-sha1", "RSA_SHA1"),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
                                "RSA_SHA256",
                            ),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha384",
                                "RSA_SHA384",
                            ),
                            (
                                "http://www.w3.org/2001/04/xmldsig-more#rsa-sha512",
                                "RSA_SHA512",
                            ),
                        ],
                        default="http://www.w3.org/2000/09/xmldsig#rsa-sha1",
                        help_text="Signature algorithm. Note that DSA_SHA1 and RSA_SHA1 are deprecated, but RSA_SHA1 is still the default value in the SAMLv2 standard. Warning: there are known issues with single-logout functionality if using anything other than SHA1 due to some hardcoded algorithm.",
                        max_length=100,
                        verbose_name="signature algorithm",
                    ),
                ),
                (
                    "digest_algorithm",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("http://www.w3.org/2000/09/xmldsig#sha1", "SHA1"),
                            ("http://www.w3.org/2001/04/xmlenc#sha256", "SHA256"),
                            ("http://www.w3.org/2001/04/xmldsig-more#sha384", "SHA384"),
                            ("http://www.w3.org/2001/04/xmlenc#sha512", "SHA512"),
                        ],
                        default="http://www.w3.org/2000/09/xmldsig#sha1",
                        help_text="Digest algorithm. Note that SHA1 is deprecated, but still the default value in the SAMLv2 standard. Warning: there are known issues with single-logout functionality if using anything other than SHA1 due to some hardcoded algorithm.",
                        max_length=100,
                        verbose_name="digest algorithm",
                    ),
                ),
                (
                    "entity_id",
                    models.CharField(
                        help_text="Service provider entity ID.",
                        max_length=100,
                        verbose_name="entity ID",
                    ),
                ),
                (
                    "base_url",
                    models.URLField(
                        help_text="Base URL of the application.",
                        max_length=100,
                        verbose_name="base URL",
                    ),
                ),
                (
                    "service_name",
                    models.CharField(
                        help_text="Name of the service you are providing.",
                        max_length=100,
                        verbose_name="service name",
                    ),
                ),
                (
                    "service_description",
                    models.CharField(
                        help_text="A description of the service you are providing.",
                        max_length=100,
                        verbose_name="service description",
                    ),
                ),
                (
                    "technical_contact_person_telephone",
                    models.CharField(
                        blank=True,
                        help_text="Telephone number of the technical person responsible for this DigiD/eHerkenning/eIDAS setup. For it to show up in the metata, you should also specify the email address.",
                        max_length=100,
                        verbose_name="technical contact: phone number",
                    ),
                ),
                (
                    "technical_contact_person_email",
                    models.CharField(
                        blank=True,
                        help_text="Email address of the technical person responsible for this DigiD/eHerkenning/eIDAS setup. For it to show up in the metadata, you should also specify the phone number.",
                        max_length=100,
                        verbose_name="technical contact: email",
                    ),
                ),
                (
                    "organization_url",
                    models.URLField(
                        blank=True,
                        help_text="URL of the organization providing the service for which DigiD/eHerkenning/eIDAS login is configured. For it to show up in the metadata, you should also specify the organization name.",
                        max_length=255,
                        verbose_name="organization URL",
                    ),
                ),
                (
                    "organization_name",
                    models.CharField(
                        blank=True,
                        help_text="URL of the organization providing the service for which DigiD/eHerkenning/eIDAS login is configured. For it to show up in the metadata, you should also specify the organization URL.",
                        max_length=100,
                        verbose_name="organization name",
                    ),
                ),
                (
                    "attribute_consuming_service_index",
                    models.CharField(
                        blank=True,
                        default="1",
                        help_text="Attribute consuming service index",
                        max_length=100,
                        verbose_name="Attribute consuming service index",
                    ),
                ),
                (
                    "requested_attributes",
                    models.JSONField(
                        default=digid_eherkenning.models.digid.default_digid_requested_attributes,
                        help_text="A list of strings (or objects) with the requested attributes, e.g. '[\"bsn\"]'",
                        verbose_name="requested attributes",
                    ),
                ),
                (
                    "slo",
                    models.BooleanField(
                        default=True,
                        help_text="If enabled, Single Logout is supported",
                        verbose_name="Single logout",
                    ),
                ),
                (
                    "certificate",
                    models.ForeignKey(
                        help_text="The private key and public certificate pair to use during the authentication flow.",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="simple_certmanager.certificate",
                        verbose_name="key pair",
                    ),
                ),
            ],
            options={
                "verbose_name": "Digid metadata configuration",
            },
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-20 08:40

from django.db import migrations, models
import privates.fields
import privates.storages


class Migration(migrations.Migration):
    dependencies = [
        (
            "digid_eherkenning",
            "0005_alter_eherkenningconfiguration_eh_service_instance_uuid_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="digidconfiguration",
            name="metadata_file_source",
            field=models.URLField(
                default="",
                help_text="The URL-source where the XML metadata file can be retrieved from.",
                max_length=255,
                verbose_name="metadata file(XML) URL",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfiguration",
            name="metadata_file_source",
            field=models.URLField(
                default="",
                help_text="The URL-source where the XML metadata file can be retrieved from.",
                max_length=255,
                verbose_name="metadata file(XML) URL",
            ),
        ),
        migrations.AlterField(
            model_name="digidconfiguration",
            name="idp_metadata_file",
            field=privates.fields.PrivateMediaFileField(
                blank=True,
                help_text="The metadata file of the identity provider. This is auto populated from the configured source URL.",
                storage=privates.storages.PrivateMediaFileSystemStorage(),
                upload_to="",
                verbose_name="identity provider metadata",
            ),
        ),
        migrations.AlterField(
            model_name="digidconfiguration",
            name="idp_service_entity_id",
            field=models.CharField(
                blank=True,
                help_text="Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note that this must match the 'entityID' attribute on the 'md:EntityDescriptor' node found in the Identity Provider's metadata. This is auto populated from the configured source URL.",
                max_length=255,
                verbose_name="identity provider service entity ID",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="idp_metadata_file",
            field=privates.fields.PrivateMediaFileField(
                blank=True,
                help_text="The metadata file of the identity provider. This is auto populated from the configured source URL.",
                storage=privates.storages.PrivateMediaFileSystemStorage(),
                upload_to="",
                verbose_name="identity provider metadata",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="idp_service_entity_id",
            field=models.CharField(
                blank=True,
                help_text="Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note that this must match the 'entityID' attribute on the 'md:EntityDescriptor' node found in the Identity Provider's metadata. This is auto populated from the configured source URL.",
                max_length=255,
                verbose_name="identity provider service entity ID",
            ),
        ),
    ]



import logging
from typing import Optional

from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth import get_user_model, logout as auth_logout
from django.contrib.auth.views import LogoutView
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import TemplateView, View

from onelogin.saml2.soap_logout_request import Soap_Logout_Request
from onelogin.saml2.utils import OneLogin_Saml2_Error, OneLogin_Saml2_ValidationError

from ..choices import DigiDAssuranceLevels, SectorType
from ..compat import get_next_page
from ..forms import SAML2Form
from ..saml2.digid import DigiDClient
from ..utils import logout_user
from .base import get_redirect_url

logger = logging.getLogger(__name__)

UserModel = get_user_model()


class DigiDLoginView(TemplateView):
    """
    DigiD - 3.3.2 - Stap 2 Authenticatievraag
    """

    template_name = "digid_eherkenning/post_binding.html"

    def get_level_of_assurance(self):
        """
        Override the default Level of Assurance (middle).

        When overriding this, remember the user has control over the request!
        """
        return DigiDAssuranceLevels.middle

    def get_relay_state(self):
        """
        TODO: It might be a good idea to sign the relay state.
        But I can't think of a way this could be abused, since
        we re-check the url when processed by the ACS.
        """
        redirect_to = self.request.GET.get("next", "")
        return get_redirect_url(self.request, redirect_to)

    #
    # TODO: It might be a good idea to change this to a post-verb.
    # I can't think of any realy attack-vectors, but seems like a good
    # idea anyways.
    #
    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        client = DigiDClient(loa=self.get_level_of_assurance())

        location, parameters = client.create_authn_request(self.request)

        context_data.update(
            {
                "url": location,
                "form": SAML2Form(
                    initial={
                        "SAMLRequest": parameters["SAMLRequest"],
                        "RelayState": self.get_relay_state(),
                    }
                ),
            }
        )
        return context_data


class DigiDAssertionConsumerServiceView(View):
    """
    DigiD - 3.3.3 Stap 5 Artifact
    """

    login_url = None
    error_messages = {
        "default": _(
            "An error occurred in the communication with DigiD. "
            "Please try again later. If this error persists, please "
            "check the website https://www.digid.nl for the latest information."
        ),
        "cancelled": _("You have cancelled logging in with DigiD."),
    }

    def get_login_url(self, **kwargs):
        url = self.get_redirect_url()
        if url:
            return url
        return resolve_url(settings.LOGIN_URL)

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url(settings.LOGIN_REDIRECT_URL)

    def get_redirect_url(self):
        redirect_to = self.request.GET.get("RelayState")
        return get_redirect_url(self.request, redirect_to)

    def get(self, request):
        errors = []
        user = auth.authenticate(
            request=request,
            digid=True,
            saml_art=request.GET.get("SAMLart"),
            errors=errors,
        )
        if user is None:
            error_code = getattr(errors[0], "code", "") if errors else ""
            error_type = (
                "cancelled"
                if error_code == OneLogin_Saml2_ValidationError.STATUS_CODE_AUTHNFAILED
                else "default"
            )
            messages.error(request, self.error_messages[error_type])
            login_url = self.get_login_url(error_type=error_type)
            return HttpResponseRedirect(login_url)

        auth.login(request, user)

        return HttpResponseRedirect(self.get_success_url())


class DigiDLogoutView(LogoutView):
    """
    Single logout with HTTP-redirect
    Local logout is done in DigidSingleLogoutCallbackView
    """

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        name_id = self.get_name_id(request)

        if not name_id:
            raise PermissionDenied(_("You are not authenticated with Digid"))

        client = DigiDClient()
        return_to = get_next_page(self)
        logout_url = client.create_logout_request(
            request, return_to=return_to, name_id=name_id
        )

        return HttpResponseRedirect(logout_url)

    @staticmethod
    def get_name_id(request) -> Optional[str]:
        """this method constructs 'name_id' using 'User.bsn' attribute"""
        # TODO perhaps it's better to use django session to store and retrieve name_id?
        bsn = getattr(request.user, "bsn", None)
        if not bsn:
            return None

        return f"{SectorType.bsn}:{request.user.bsn}"


class DigidSingleLogoutRedirectView(View):
    """
    Logout response from IdP when SP initiates logout (step U5) with HTTP Redirect binding
    """

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url(settings.LOGOUT_REDIRECT_URL)

    def get_redirect_url(self):
        redirect_to = self.request.GET.get("RelayState")
        return get_redirect_url(self.request, redirect_to)

    def get(self, request, *args, **kwargs):
        user = request.user
        client = DigiDClient()
        try:
            client.handle_logout_response(
                request,
                keep_local_session=False,
                delete_session_cb=lambda: auth_logout(request),
            )
        except OneLogin_Saml2_Error as e:
            error_message = "An error occurred during logout from Digid"
            logger.error("%s: %s", error_message, e.args[0])
            messages.error(request, _(error_message))
        else:
            logger.info("User %s has successfully logged out of Digid", user)

        return HttpResponseRedirect(self.get_success_url())


@method_decorator(csrf_exempt, name="dispatch")
class DigidSingleLogoutSoapView(View):
    """
    Logout request from IdP when Idp initiates logout (step U3) with SOAP binding
    """

    def post(self, request, *args, **kwargs):
        """handle Logout Response with SOAP binding (step U3)"""

        logger.info(
            "Received Logout Request (POST) from IdP: request body=%s", request.body
        )

        client = DigiDClient()
        logout_response = client.handle_logout_request(
            request,
            keep_local_session=False,
            delete_session_cb=lambda: self.logout_digid_user(request),
        )
        # SAML binding, section 3.2.3.3
        status_code = 500 if "faultcode" in logout_response else 200

        return HttpResponse(
            logout_response, status=status_code, content_type="text/xml"
        )

    @staticmethod
    def logout_digid_user(request):
        """
        delete all sessions with the user identified in nameId of logout request
        """
        name_id = Soap_Logout_Request(request=request.body, settings={}).get_name_id()
        sector_code, bsn = name_id.split(":")

        try:
            user = UserModel.digid_objects.get_by_bsn(bsn)
        except UserModel.DoesNotExist:
            logger.error(
                "User with BSN %s doesn't exist and therefore can't be logged out", bsn
            )
            return

        # delete all user sessions
        try:
            logout_user(user)
        except RuntimeError:
            logger.error(
                "The error occurred during forceful logout of User %s. "
                "Check if 'sessionprofile' app is added into INSTALLED_APPS",
                user,
            )

        logger.info("User %s has been forcefully logged out of Digid", user)



from django.urls import reverse

import pytest

from digid_eherkenning.models import (
    DigidConfiguration,
    EherkenningConfiguration,
    eherkenning,
)

pytestmark = pytest.mark.django_db


def test_digid_metadata_properly_displayed(digid_config, client):
    response = client.get(reverse("metadata:digid"))

    assert response.status_code == 200


def test_digid_metadata_not_properly_displayed(digid_config, client):
    DigidConfiguration.get_solo().delete()

    response = client.get(reverse("metadata:digid"))

    assert response.status_code == 400


def test_eherkenning_metadata_properly_displayed(eherkenning_config, client):
    response = client.get(reverse("metadata:eherkenning"))

    assert response.status_code == 200


def test_eherkenning_metadata_not_properly_displayed(eherkenning_config, client):
    EherkenningConfiguration.get_solo().delete()

    response = client.get(reverse("metadata:eherkenning"))

    assert response.status_code == 400


def test_eherkenning_metadata_not_properly_displayed_without_validation_errors(
    eherkenning_config, client
):
    # make some blank=False fields empty, which passed config.clean()
    eherkenning_config.entity_id = ""
    eherkenning_config.makelaar_id = ""
    eherkenning_config.oin = ""
    eherkenning_config.save()

    response = client.get(reverse("metadata:eherkenning"))

    assert response.status_code == 400


def test_dienstcatalogus_metadata_properly_displayed(eherkenning_config, client):
    eherkenning_config.makelaar_id = "00000000000000000022"
    eherkenning_config.save()

    response = client.get(reverse("metadata:eh-dienstcatalogus"))

    assert response.status_code == 200


def test_dienstcatalogus_metadata_not_properly_displayed(eherkenning_config, client):
    EherkenningConfiguration.get_solo().delete()

    response = client.get(reverse("metadata:eh-dienstcatalogus"))

    assert response.status_code == 400



from django.utils import timezone

from ...models import DigidConfiguration
from ...saml2.digid import generate_digid_metadata
from ._base import SamlMetadataBaseCommand

try:
    from argparse import BooleanOptionalAction
except ImportError:
    from ..utils import BooleanOptionalAction


class Command(SamlMetadataBaseCommand):
    help = "Create the DigiD metadata file"
    config_model = DigidConfiguration
    default_certificate_label = "DigiD"

    def add_arguments(self, parser):
        super().add_arguments(parser)

        config: DigidConfiguration = self._get_config()

        parser.add_argument(
            "--slo",
            default=config.slo,
            action=BooleanOptionalAction,
            help="If '--slo' is present, Single Logout is supported. To turn it off use '--no-slo'",
        )
        parser.add_argument(
            "--attribute-consuming-service-index",
            type=str,
            help="Attribute consuming service index, defaults to 1",
            default="1",
        )

    def get_filename(self):
        date_string = timezone.now().date().isoformat()
        return f"digid-metadata-{date_string}.xml"

    def generate_metadata(self, options):
        return generate_digid_metadata()



import pytest

from digid_eherkenning.mock.idp.forms import PasswordLoginForm


@pytest.mark.parametrize(
    "auth_name, auth_pass, name_has_error, pass_has_error",
    [
        ("296648875", "password", False, False),  # OK
        ("abcdefghe", "password", True, False),  # bsn wrong type
        ("2966488759", "password", True, False),  # bsn too long
        ("29664887", "password", True, False),  # bsn too short
        ("123456789", "password", True, False),  # bsn wrong checksum
        ("296648875", "", False, True),  # missing password
    ],
)
def test_password_login_form_validate(
    auth_name, auth_pass, name_has_error, pass_has_error
):
    form = PasswordLoginForm(data={"auth_name": auth_name, "auth_pass": auth_pass})

    assert form.has_error("auth_name") is name_has_error
    assert form.has_error("auth_pass") is pass_has_error
    assert form.is_valid() is not (name_has_error or pass_has_error)



from django.db import IntegrityError

import pytest

from digid_eherkenning.choices import AssuranceLevels


@pytest.mark.django_db()
def test_fixing_misconfigured_eherkenning(migrator):
    old_state = migrator.apply_initial_migration(
        ("digid_eherkenning", "0003_digidconfiguration_artifact_resolve_content_type")
    )
    OldEherkenningConfiguration = old_state.apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )

    # misconfigured entry; attribute had default, but also blank=True
    OldEherkenningConfiguration(loa="").save()

    new_state = migrator.apply_tested_migration(
        ("digid_eherkenning", "0004_alter_eherkenningconfiguration_loa")
    )
    EherkenningConfiguration = new_state.apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )
    config = EherkenningConfiguration.objects.get()
    assert config.loa == AssuranceLevels.low_plus

    # impossible to misconfigure
    config.loa = ""
    with pytest.raises(IntegrityError):
        config.save()



from pathlib import Path

from django.conf import settings

BASE_DIR = Path(__file__).parent.resolve()

EHERKENNING_DS_XSD = str(BASE_DIR / "xsd" / "eherkenning-dc.xml")

empty = object()


# Public settings
class Defaults:
    DIGID_SESSION_AGE: int = 60 * 15  # 15 minutes, in seconds


def get_setting(name: str):
    """
    Get the runtime setting value or use the default value if not specified.
    """
    value = getattr(settings, name, empty)
    if value is not empty:
        return value

    default = getattr(Defaults, name)
    return default



# Generated by Django 4.2.4 on 2023-09-01 09:23

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("digid_eherkenning", "0004_alter_eherkenningconfiguration_loa"),
    ]

    operations = [
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="eh_service_instance_uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text="UUID of the eHerkenning service instance. Once entered into catalogues, changing the value is a manual process.",
                verbose_name="eHerkenning service instance UUID",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="eidas_service_instance_uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text="UUID of the eIDAS service instance. Once entered into catalogues, changing the value is a manual process.",
                verbose_name="eIDAS service instance UUID",
            ),
        ),
    ]



from io import StringIO

from django.core.management import CommandError, call_command
from django.test import TestCase

import pytest
from lxml import etree

from digid_eherkenning.choices import AssuranceLevels
from digid_eherkenning.models import EherkenningConfiguration
from digid_eherkenning.saml2.eherkenning import (
    create_service_catalogus,
    generate_dienst_catalogus_metadata,
)

from .conftest import EHERKENNING_TEST_CERTIFICATE_FILE, EHERKENNING_TEST_KEY_FILE
from .mixins import EherkenningMetadataMixin

NAMESPACES = {
    "esc": "urn:etoegang:1.13:service-catalog",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
    "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
}


def _xpath(node, xpath: str) -> str:
    needle = node.find(xpath, namespaces=NAMESPACES)
    return needle.text


@pytest.mark.django_db
def test_wants_assertions_signed_setting_default(
    eherkenning_config_defaults, temp_private_root
):
    config = eherkenning_config_defaults
    config.eh_service_uuid = "005f18b8-0114-4a1d-963a-ee8e80a08f3f"
    config.eh_service_instance_uuid = "54efe0fe-c1a7-42da-9612-d84bf3c8fb07"
    config.eidas_service_uuid = "2e167de1-8bef-4d5a-ab48-8fa020e9e631"
    config.eidas_service_instance_uuid = "9ba1b0ee-c0d3-437e-87ac-f577098c7e15"
    config.oin = "00000000000000000000"
    config.makelaar_id = "00000000000000000000"
    config.service_name = "Example"
    config.service_description = "Description"
    config.eh_attribute_consuming_service_index = "1"
    config.eidas_attribute_consuming_service_index = "2"
    config.save()

    conf = config.as_dict()

    catalogus = create_service_catalogus(conf, validate=False)

    # Parse XML
    tree = etree.XML(catalogus)

    # Test that there are 2 ServiceDefinition and 2 ServiceInstance nodes inside the ServiceProvider
    service_provider_node = tree.find(".//esc:ServiceProvider", namespaces=NAMESPACES)

    assert (
        service_provider_node.find(".//esc:ServiceProviderID", namespaces=NAMESPACES)
        is not None
    )
    assert (
        service_provider_node.find(
            ".//esc:OrganizationDisplayName", namespaces=NAMESPACES
        )
        is not None
    )

    service_definition_nodes = service_provider_node.findall(
        ".//esc:ServiceDefinition", namespaces=NAMESPACES
    )

    assert len(service_definition_nodes) == 2

    service_node_0 = service_definition_nodes[0]
    assert (
        _xpath(service_node_0, ".//esc:ServiceUUID")
        == "005f18b8-0114-4a1d-963a-ee8e80a08f3f"
    )
    assert _xpath(service_node_0, ".//esc:ServiceName") == "Example"
    assert _xpath(service_node_0, ".//esc:ServiceDescription") == "Description"
    assert (
        _xpath(service_node_0, ".//esc:EntityConcernedTypesAllowed")
        == "urn:etoegang:1.9:EntityConcernedID:RSIN"
    )
    assert (
        _xpath(service_node_0, ".//esc:ServiceRestrictionsAllowed")
        == "urn:etoegang:1.9:ServiceRestriction:Vestigingsnr"
    )

    service_node_1 = service_definition_nodes[1]
    assert (
        _xpath(service_node_1, ".//esc:ServiceUUID")
        == "2e167de1-8bef-4d5a-ab48-8fa020e9e631"
    )
    assert _xpath(service_node_1, ".//esc:ServiceName") == "Example (eIDAS)"
    assert _xpath(service_node_1, ".//esc:ServiceDescription") == "Description"
    assert (
        _xpath(service_node_1, ".//esc:EntityConcernedTypesAllowed")
        == "urn:etoegang:1.9:EntityConcernedID:Pseudo"
    )

    service_instance_nodes = service_provider_node.findall(
        ".//esc:ServiceInstance", namespaces=NAMESPACES
    )

    assert len(service_instance_nodes) == 2
    instance_node_0 = service_instance_nodes[0]
    assert (
        _xpath(instance_node_0, ".//esc:ServiceID")
        == "urn:etoegang:DV:00000000000000000000:services:1"
    )
    assert (
        _xpath(instance_node_0, ".//esc:ServiceUUID")
        == "54efe0fe-c1a7-42da-9612-d84bf3c8fb07"
    )
    assert (
        _xpath(instance_node_0, ".//esc:InstanceOfService")
        == "005f18b8-0114-4a1d-963a-ee8e80a08f3f"
    )
    assert (
        instance_node_0.find(".//esc:ServiceCertificate", namespaces=NAMESPACES).find(
            ".//md:KeyDescriptor", namespaces=NAMESPACES
        )
    ) is not None
    assert (instance_node_0.find(".//esc:Classifiers", namespaces=NAMESPACES)) is None

    service_url_nodes = instance_node_0.findall(
        ".//esc:ServiceURL", namespaces=NAMESPACES
    )
    assert len(service_url_nodes) == 2
    assert (
        service_url_nodes[0].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "nl"
    )
    assert (
        service_url_nodes[1].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "en"
    )

    privacy_policy_nodes = instance_node_0.findall(
        ".//esc:PrivacyPolicyURL", namespaces=NAMESPACES
    )
    assert len(privacy_policy_nodes) == 2
    assert (
        privacy_policy_nodes[0].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "nl"
    )
    assert (
        privacy_policy_nodes[1].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "en"
    )

    instance_node_1 = service_instance_nodes[1]
    assert (
        _xpath(instance_node_1, ".//esc:ServiceID")
        == "urn:etoegang:DV:00000000000000000000:services:2"
    )
    assert (
        _xpath(instance_node_1, ".//esc:ServiceUUID")
        == "9ba1b0ee-c0d3-437e-87ac-f577098c7e15"
    )
    assert (
        _xpath(instance_node_1, ".//esc:InstanceOfService")
        == "2e167de1-8bef-4d5a-ab48-8fa020e9e631"
    )
    assert (
        instance_node_1.find(".//esc:ServiceCertificate", namespaces=NAMESPACES).find(
            ".//md:KeyDescriptor", namespaces=NAMESPACES
        )
    ) is not None
    assert (
        instance_node_1.find(".//esc:Classifiers", namespaces=NAMESPACES)
    ) is not None


@pytest.mark.django_db
def test_catalogus_with_requested_attributes_with_purpose_statement(
    eherkenning_config_defaults, temp_private_root
):
    config = eherkenning_config_defaults
    config.oin = "00000000000000000000"
    config.makelaar_id = "1" * 20
    config.eh_requested_attributes = [
        {
            "name": "Test Attribute",
            "required": False,
            "purpose_statements": {
                "nl": "Voor testen.",
                "en": "For testing.",
            },
        }
    ]
    config.eidas_requested_attributes = []
    config.save()

    conf = config.as_dict()
    catalogus = create_service_catalogus(conf)

    tree = etree.XML(catalogus)

    requested_attribute_nodes = tree.findall(
        ".//esc:RequestedAttribute", namespaces=NAMESPACES
    )
    assert len(requested_attribute_nodes) == 1

    purpose_statement_nodes = tree.findall(
        ".//esc:PurposeStatement", namespaces=NAMESPACES
    )
    assert len(purpose_statement_nodes) == 2
    assert purpose_statement_nodes[0].text == "Voor testen."
    assert purpose_statement_nodes[1].text == "For testing."
    assert (
        purpose_statement_nodes[0].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "nl"
    )
    assert (
        purpose_statement_nodes[1].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "en"
    )


@pytest.mark.django_db
def test_catalogus_with_requested_attributes_without_purpose_statement(
    eherkenning_config_defaults, temp_private_root
):
    config = eherkenning_config_defaults
    config.oin = "00000000000000000000"
    config.makelaar_id = "1" * 20
    config.eh_requested_attributes = [
        {
            "name": "Test Attribute",
            "required": False,
        }
    ]
    config.eidas_requested_attributes = []
    config.save()

    conf = config.as_dict()
    catalogus = create_service_catalogus(conf)

    # FIXME: incorporate this in DB config?
    conf["services"][0]["service_name"] = {
        "nl": "Voorbeeld dienst",
        "en": "Example service",
    }

    catalogus = create_service_catalogus(conf)

    tree = etree.XML(catalogus)

    requested_attribute_nodes = tree.findall(
        ".//esc:RequestedAttribute", namespaces=NAMESPACES
    )
    assert len(requested_attribute_nodes) == 1

    purpose_statement_nodes = tree.findall(
        ".//esc:PurposeStatement", namespaces=NAMESPACES
    )
    assert len(purpose_statement_nodes) == 2
    assert purpose_statement_nodes[0].text == "Voorbeeld dienst"
    assert purpose_statement_nodes[1].text == "Example service"
    assert (
        purpose_statement_nodes[0].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "nl"
    )
    assert (
        purpose_statement_nodes[1].attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        == "en"
    )


@pytest.mark.django_db
def test_makelaar_oin_is_configurable(eherkenning_config_defaults, temp_private_root):
    config = EherkenningConfiguration.get_solo()
    config.organisation_name = "Example"
    config.service_name = "Example"
    config.oin = "00000000000000000000"
    config.makelaar_id = "00000000000000000123"
    config.save()
    conf = config.as_dict()

    catalogus = create_service_catalogus(conf, validate=False)

    # Parse XML
    tree = etree.XML(catalogus)

    makelaar_id_nodes = tree.findall(
        ".//esc:HerkenningsmakelaarId", namespaces=NAMESPACES
    )
    for node in makelaar_id_nodes:
        assert node.text == "00000000000000000123"


@pytest.mark.usefixtures("eherkenning_config_defaults", "temp_private_root")
class DienstCatalogusMetadataTests(EherkenningMetadataMixin, TestCase):
    def test_generate_metadata_all_options_specified(self):
        self.eherkenning_config.signature_algorithm = (
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
        )
        self.eherkenning_config.digest_algorithm = (
            "http://www.w3.org/2001/04/xmlenc#sha256"
        )
        self.eherkenning_config.technical_contact_person_telephone = "06123123123"
        self.eherkenning_config.technical_contact_person_email = "test@test.nl"
        self.eherkenning_config.organization_name = "Test Organisation"
        self.eherkenning_config.organization_url = "http://test-organisation.nl"
        self.eherkenning_config.service_description_url = (
            "http://test-organisation.nl/service-description/"
        )
        self.eherkenning_config.eh_loa = AssuranceLevels.high
        self.eherkenning_config.eidas_loa = AssuranceLevels.low
        self.eherkenning_config.eidas_requested_attributes = [
            {
                "name": "urn:etoegang:1.9:attribute:FirstName",
                "required": True,
                "purpose_statements": {
                    "en": "A reason for the first name",
                    "nl": "Een reden voor de voornaam",
                },
            },
            {
                "name": "urn:etoegang:1.9:attribute:FamilyName",
                "required": True,
                "purpose_statements": {
                    "en": "A reason for the last name",
                    "nl": "Een reden voor de achternaam",
                },
            },
            {
                "name": "urn:etoegang:1.9:attribute:DateOfBirth",
                "required": True,
                "purpose_statements": {
                    "en": "A reason for the date of birth",
                    "nl": "Een reden voor de geboortedatum",
                },
            },
            {
                "name": "urn:etoegang:1.11:attribute-represented:CompanyName",
                "required": True,
                "purpose_statements": {
                    "en": "A reason for the company name",
                    "nl": "Een reden voor de bedrijfnaam",
                },
            },
        ]
        self.eherkenning_config.save()

        eherkenning_dienstcatalogus_metadata = generate_dienst_catalogus_metadata(
            self.eherkenning_config
        )
        self.assertEqual(eherkenning_dienstcatalogus_metadata[:5], b"<?xml")
        service_catalogue_node = etree.XML(eherkenning_dienstcatalogus_metadata)

        signature_algorithm_node = service_catalogue_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = service_catalogue_node.find(
            ".//ds:DigestMethod",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        # Service Provider
        service_provider_id_node = service_catalogue_node.find(
            ".//esc:ServiceProviderID",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "00000000000000000011",
            service_provider_id_node.text,
        )

        oganisation_display_node = service_catalogue_node.find(
            ".//esc:OrganizationDisplayName",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "Test Organisation",
            oganisation_display_node.text,
        )

        # Services
        service_definition_nodes = service_catalogue_node.findall(
            ".//esc:ServiceDefinition",
            namespaces=NAMESPACES,
        )
        self.assertEqual(2, len(service_definition_nodes))

        eherkenning_definition_node, eidas_definition_node = service_definition_nodes

        # eHerkenning service definition
        uuid_node = eherkenning_definition_node.find(
            ".//esc:ServiceUUID",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(uuid_node)

        service_name_node = eherkenning_definition_node.find(
            ".//esc:ServiceName",
            namespaces=NAMESPACES,
        )
        self.assertEqual("Test Service Name", service_name_node.text)

        service_description_node = eherkenning_definition_node.find(
            ".//esc:ServiceDescription",
            namespaces=NAMESPACES,
        )
        self.assertEqual("Test Service Description", service_description_node.text)
        service_description_url_node = eherkenning_definition_node.find(
            ".//esc:ServiceDescriptionURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "http://test-organisation.nl/service-description/",
            service_description_url_node.text,
        )

        loa_node = eherkenning_definition_node.find(
            ".//saml:AuthnContextClassRef",
            namespaces=NAMESPACES,
        )
        self.assertEqual("urn:etoegang:core:assurance-class:loa4", loa_node.text)

        makelaar_id_node = eherkenning_definition_node.find(
            ".//esc:HerkenningsmakelaarId",
            namespaces=NAMESPACES,
        )
        self.assertEqual("00000000000000000022", makelaar_id_node.text)

        entity_concerned_nodes = eherkenning_definition_node.findall(
            ".//esc:EntityConcernedTypesAllowed",
            namespaces=NAMESPACES,
        )
        self.assertEqual(3, len(entity_concerned_nodes))
        self.assertEqual("1", entity_concerned_nodes[0].attrib["setNumber"])
        self.assertEqual(
            "urn:etoegang:1.9:EntityConcernedID:RSIN", entity_concerned_nodes[0].text
        )
        self.assertEqual("1", entity_concerned_nodes[1].attrib["setNumber"])
        self.assertEqual(
            "urn:etoegang:1.9:EntityConcernedID:KvKnr", entity_concerned_nodes[1].text
        )
        self.assertEqual("2", entity_concerned_nodes[2].attrib["setNumber"])
        self.assertEqual(
            "urn:etoegang:1.9:EntityConcernedID:KvKnr", entity_concerned_nodes[2].text
        )

        # eIDAS service definition
        uuid_node = eidas_definition_node.find(
            ".//esc:ServiceUUID",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(uuid_node)

        service_name_node = eidas_definition_node.find(
            ".//esc:ServiceName",
            namespaces=NAMESPACES,
        )
        self.assertEqual("Test Service Name (eIDAS)", service_name_node.text)

        service_description_node = eidas_definition_node.find(
            ".//esc:ServiceDescription",
            namespaces=NAMESPACES,
        )
        self.assertEqual("Test Service Description", service_description_node.text)

        loa_node = eidas_definition_node.find(
            ".//saml:AuthnContextClassRef",
            namespaces=NAMESPACES,
        )
        self.assertEqual("urn:etoegang:core:assurance-class:loa2", loa_node.text)

        makelaar_id_node = eidas_definition_node.find(
            ".//esc:HerkenningsmakelaarId",
            namespaces=NAMESPACES,
        )
        self.assertEqual("00000000000000000022", makelaar_id_node.text)

        entity_concerned_nodes = eidas_definition_node.findall(
            ".//esc:EntityConcernedTypesAllowed",
            namespaces=NAMESPACES,
        )
        self.assertEqual(1, len(entity_concerned_nodes))
        self.assertEqual(
            "urn:etoegang:1.9:EntityConcernedID:Pseudo", entity_concerned_nodes[0].text
        )

        requested_attribute_nodes = eidas_definition_node.findall(
            ".//esc:RequestedAttribute", namespaces=NAMESPACES
        )
        self.assertEqual(len(requested_attribute_nodes), 4)
        self.assertEqual(
            requested_attribute_nodes[0].attrib["Name"],
            "urn:etoegang:1.9:attribute:FirstName",
        )
        self.assertEqual(requested_attribute_nodes[0].attrib["isRequired"], "true")
        self.assertEqual(
            requested_attribute_nodes[1].attrib["Name"],
            "urn:etoegang:1.9:attribute:FamilyName",
        )
        self.assertEqual(requested_attribute_nodes[1].attrib["isRequired"], "true")
        self.assertEqual(
            requested_attribute_nodes[2].attrib["Name"],
            "urn:etoegang:1.9:attribute:DateOfBirth",
        )
        self.assertEqual(requested_attribute_nodes[2].attrib["isRequired"], "true")
        self.assertEqual(
            requested_attribute_nodes[3].attrib["Name"],
            "urn:etoegang:1.11:attribute-represented:CompanyName",
        )
        self.assertEqual(requested_attribute_nodes[3].attrib["isRequired"], "true")

        # Service instances
        service_instance_nodes = service_catalogue_node.findall(
            ".//esc:ServiceInstance",
            namespaces=NAMESPACES,
        )
        self.assertEqual(2, len(service_instance_nodes))

        eherkenning_instance_node, eidas_instance_node = service_instance_nodes

        # Service instance eHerkenning
        service_id_node = eherkenning_instance_node.find(
            ".//esc:ServiceID",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9050", service_id_node.text
        )

        service_url_node = eherkenning_instance_node.find(
            ".//esc:ServiceURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-entity.id", service_url_node.text)

        privacy_url_node = eherkenning_instance_node.find(
            ".//esc:PrivacyPolicyURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-privacy.nl", privacy_url_node.text)

        makelaar_id_node = eherkenning_instance_node.find(
            ".//esc:HerkenningsmakelaarId",
            namespaces=NAMESPACES,
        )
        self.assertEqual("00000000000000000022", makelaar_id_node.text)

        key_name_node = eherkenning_instance_node.find(
            ".//ds:KeyName",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(key_name_node)
        certificate_node = eherkenning_instance_node.find(
            ".//ds:X509Certificate",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(certificate_node)

        classifier_node = eherkenning_instance_node.findall(
            ".//esc:Classifier",
            namespaces=NAMESPACES,
        )
        self.assertEqual(0, len(classifier_node))

        # Service instance eIDAS
        service_id_node = eidas_instance_node.find(
            ".//esc:ServiceID",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9051", service_id_node.text
        )

        service_url_node = eidas_instance_node.find(
            ".//esc:ServiceURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-entity.id", service_url_node.text)

        privacy_url_node = eidas_instance_node.find(
            ".//esc:PrivacyPolicyURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-privacy.nl", privacy_url_node.text)

        makelaar_id_node = eidas_instance_node.find(
            ".//esc:HerkenningsmakelaarId",
            namespaces=NAMESPACES,
        )
        self.assertEqual("00000000000000000022", makelaar_id_node.text)

        key_name_node = eidas_instance_node.find(
            ".//ds:KeyName",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(key_name_node)
        certificate_node = eidas_instance_node.find(
            ".//ds:X509Certificate",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(certificate_node)

        classifier_node = eidas_instance_node.findall(
            ".//esc:Classifier",
            namespaces=NAMESPACES,
        )
        self.assertEqual(1, len(classifier_node))
        self.assertEqual("eIDAS-inbound", classifier_node[0].text)

    def test_no_eidas_service(self):
        self.eherkenning_config.no_eidas = True
        self.eherkenning_config.save()

        eherkenning_dienstcatalogus_metadata = generate_dienst_catalogus_metadata(
            self.eherkenning_config
        )
        service_catalogue_node = etree.XML(eherkenning_dienstcatalogus_metadata)

        service_instance_nodes = service_catalogue_node.findall(
            ".//esc:ServiceInstance",
            namespaces=NAMESPACES,
        )
        self.assertEqual(1, len(service_instance_nodes))

        eherkenning_instance_node = service_instance_nodes[0]

        # Service instance eHerkenning
        service_id_node = eherkenning_instance_node.find(
            ".//esc:ServiceID",
            namespaces=NAMESPACES,
        )
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9050", service_id_node.text
        )

        service_url_node = eherkenning_instance_node.find(
            ".//esc:ServiceURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-entity.id", service_url_node.text)

        privacy_url_node = eherkenning_instance_node.find(
            ".//esc:PrivacyPolicyURL",
            namespaces=NAMESPACES,
        )
        self.assertEqual("http://test-privacy.nl", privacy_url_node.text)

        makelaar_id_node = eherkenning_instance_node.find(
            ".//esc:HerkenningsmakelaarId",
            namespaces=NAMESPACES,
        )
        self.assertEqual("00000000000000000022", makelaar_id_node.text)

        key_name_node = eherkenning_instance_node.find(
            ".//ds:KeyName",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(key_name_node)
        certificate_node = eherkenning_instance_node.find(
            ".//ds:X509Certificate",
            namespaces=NAMESPACES,
        )
        self.assertIsNotNone(certificate_node)

        classifier_node = eherkenning_instance_node.findall(
            ".//esc:Classifier",
            namespaces=NAMESPACES,
        )
        self.assertEqual(0, len(classifier_node))


@pytest.mark.django_db
def test_generate_metadata_all_options_specified(temp_private_root):
    stdout = StringIO()

    call_command(
        "generate_eherkenning_dienstcatalogus",
        "--no-save-config",
        stdout=stdout,
        key_file=str(EHERKENNING_TEST_KEY_FILE),
        cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
        signature_algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
        digest_algorithm="http://www.w3.org/2001/04/xmlenc#sha256",
        entity_id="http://test-entity.id",
        base_url="http://test-entity.id",
        organization_name="Test Organisation",
        eh_attribute_consuming_service_index="9050",
        eidas_attribute_consuming_service_index="9051",
        oin="00000001112223330000",
        service_name="Test Service Name",
        service_description="Test Service Description",
        makelaar_id="00000003332221110000",
        privacy_policy="http://test-privacy.nl",
        test=True,
    )

    output = stdout.getvalue()
    service_catalogue_node = etree.XML(output.encode("utf-8"))

    signature_algorithm_node = service_catalogue_node.find(
        ".//ds:SignatureMethod",
        namespaces=NAMESPACES,
    )
    assert (
        signature_algorithm_node.attrib["Algorithm"]
        == "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
    )

    digest_algorithm_node = service_catalogue_node.find(
        ".//ds:DigestMethod",
        namespaces=NAMESPACES,
    )
    assert (
        digest_algorithm_node.attrib["Algorithm"]
        == "http://www.w3.org/2001/04/xmlenc#sha256"
    )

    # Service Provider
    service_provider_id_node = service_catalogue_node.find(
        ".//esc:ServiceProviderID",
        namespaces=NAMESPACES,
    )
    assert service_provider_id_node.text == "00000001112223330000"

    oganisation_display_node = service_catalogue_node.find(
        ".//esc:OrganizationDisplayName",
        namespaces=NAMESPACES,
    )
    assert oganisation_display_node.text == "Test Organisation"

    # Services
    service_definition_nodes = service_catalogue_node.findall(
        ".//esc:ServiceDefinition",
        namespaces=NAMESPACES,
    )
    assert len(service_definition_nodes) == 2

    eherkenning_definition_node, eidas_definition_node = service_definition_nodes

    # eHerkenning service definition
    uuid_node = eherkenning_definition_node.find(
        ".//esc:ServiceUUID",
        namespaces=NAMESPACES,
    )
    assert uuid_node is not None

    service_name_node = eherkenning_definition_node.find(
        ".//esc:ServiceName",
        namespaces=NAMESPACES,
    )
    assert service_name_node.text == "Test Service Name"

    service_description_node = eherkenning_definition_node.find(
        ".//esc:ServiceDescription",
        namespaces=NAMESPACES,
    )
    assert service_description_node.text == "Test Service Description"

    loa_node = eherkenning_definition_node.find(
        ".//saml:AuthnContextClassRef",
        namespaces=NAMESPACES,
    )
    assert loa_node.text == "urn:etoegang:core:assurance-class:loa3"

    makelaar_id_node = eherkenning_definition_node.find(
        ".//esc:HerkenningsmakelaarId",
        namespaces=NAMESPACES,
    )
    assert makelaar_id_node.text == "00000003332221110000"

    entity_concerned_nodes = eherkenning_definition_node.findall(
        ".//esc:EntityConcernedTypesAllowed",
        namespaces=NAMESPACES,
    )
    assert len(entity_concerned_nodes) == 3
    assert entity_concerned_nodes[0].attrib["setNumber"] == "1"
    assert entity_concerned_nodes[0].text == "urn:etoegang:1.9:EntityConcernedID:RSIN"
    assert entity_concerned_nodes[1].attrib["setNumber"] == "1"
    assert entity_concerned_nodes[1].text == "urn:etoegang:1.9:EntityConcernedID:KvKnr"
    assert entity_concerned_nodes[2].attrib["setNumber"] == "2"
    assert entity_concerned_nodes[2].text == "urn:etoegang:1.9:EntityConcernedID:KvKnr"

    # eIDAS service definition
    uuid_node = eidas_definition_node.find(
        ".//esc:ServiceUUID",
        namespaces=NAMESPACES,
    )
    assert uuid_node is not None

    service_name_node = eidas_definition_node.find(
        ".//esc:ServiceName",
        namespaces=NAMESPACES,
    )
    assert service_name_node.text == "Test Service Name (eIDAS)"

    service_description_node = eidas_definition_node.find(
        ".//esc:ServiceDescription",
        namespaces=NAMESPACES,
    )
    assert service_description_node.text == "Test Service Description"

    loa_node = eidas_definition_node.find(
        ".//saml:AuthnContextClassRef",
        namespaces=NAMESPACES,
    )
    assert loa_node.text == "urn:etoegang:core:assurance-class:loa3"

    makelaar_id_node = eidas_definition_node.find(
        ".//esc:HerkenningsmakelaarId",
        namespaces=NAMESPACES,
    )
    assert makelaar_id_node.text == "00000003332221110000"

    entity_concerned_nodes = eidas_definition_node.findall(
        ".//esc:EntityConcernedTypesAllowed",
        namespaces=NAMESPACES,
    )
    assert len(entity_concerned_nodes) == 1
    assert entity_concerned_nodes[0].text == "urn:etoegang:1.9:EntityConcernedID:Pseudo"

    # Service instances
    service_instance_nodes = service_catalogue_node.findall(
        ".//esc:ServiceInstance",
        namespaces=NAMESPACES,
    )
    assert len(service_instance_nodes) == 2

    eherkenning_instance_node, eidas_instance_node = service_instance_nodes

    # Service instance eHerkenning
    service_id_node = eherkenning_instance_node.find(
        ".//esc:ServiceID",
        namespaces=NAMESPACES,
    )
    assert service_id_node.text == "urn:etoegang:DV:00000001112223330000:services:9050"

    service_url_node = eherkenning_instance_node.find(
        ".//esc:ServiceURL",
        namespaces=NAMESPACES,
    )
    assert service_url_node.text == "http://test-entity.id"

    privacy_url_node = eherkenning_instance_node.find(
        ".//esc:PrivacyPolicyURL",
        namespaces=NAMESPACES,
    )
    assert privacy_url_node.text == "http://test-privacy.nl"

    makelaar_id_node = eherkenning_instance_node.find(
        ".//esc:HerkenningsmakelaarId",
        namespaces=NAMESPACES,
    )
    assert makelaar_id_node.text == "00000003332221110000"

    key_name_node = eherkenning_instance_node.find(
        ".//ds:KeyName",
        namespaces=NAMESPACES,
    )
    assert key_name_node is not None
    certificate_node = eherkenning_instance_node.find(
        ".//ds:X509Certificate",
        namespaces=NAMESPACES,
    )
    assert certificate_node is not None

    classifier_node = eherkenning_instance_node.findall(
        ".//esc:Classifier",
        namespaces=NAMESPACES,
    )
    assert len(classifier_node) == 0

    # Service instance eIDAS
    service_id_node = eidas_instance_node.find(
        ".//esc:ServiceID",
        namespaces=NAMESPACES,
    )
    assert service_id_node.text == "urn:etoegang:DV:00000001112223330000:services:9051"

    service_url_node = eidas_instance_node.find(
        ".//esc:ServiceURL",
        namespaces=NAMESPACES,
    )
    assert service_url_node.text == "http://test-entity.id"

    privacy_url_node = eidas_instance_node.find(
        ".//esc:PrivacyPolicyURL",
        namespaces=NAMESPACES,
    )
    assert privacy_url_node.text == "http://test-privacy.nl"

    makelaar_id_node = eidas_instance_node.find(
        ".//esc:HerkenningsmakelaarId",
        namespaces=NAMESPACES,
    )
    assert makelaar_id_node.text == "00000003332221110000"

    key_name_node = eidas_instance_node.find(
        ".//ds:KeyName",
        namespaces=NAMESPACES,
    )
    assert key_name_node is not None
    certificate_node = eidas_instance_node.find(
        ".//ds:X509Certificate",
        namespaces=NAMESPACES,
    )
    assert certificate_node is not None

    classifier_node = eidas_instance_node.findall(
        ".//esc:Classifier",
        namespaces=NAMESPACES,
    )
    assert len(classifier_node) == 1
    assert classifier_node[0].text == "eIDAS-inbound"


@pytest.mark.django_db
def test_missing_required_properties():
    with pytest.raises(CommandError):
        call_command("generate_eherkenning_dienstcatalogus")


@pytest.mark.django_db
def test_no_eidas_service(temp_private_root):
    stdout = StringIO()

    call_command(
        "generate_eherkenning_dienstcatalogus",
        "--no-save-config",
        stdout=stdout,
        key_file=str(EHERKENNING_TEST_KEY_FILE),
        cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
        signature_algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
        digest_algorithm="http://www.w3.org/2001/04/xmlenc#sha256",
        entity_id="http://test-entity.id",
        base_url="http://test-entity.id",
        organization_name="Test Organisation",
        eh_attribute_consuming_service_index="9050",
        no_eidas=True,
        oin="00000001112223330000",
        service_name="Test Service Name",
        service_description="Test Service Description",
        makelaar_id="00000003332221110000",
        privacy_policy="http://test-privacy.nl",
        test=True,
    )

    output = stdout.getvalue()
    service_catalogue_node = etree.XML(output.encode("utf-8"))

    service_instance_nodes = service_catalogue_node.findall(
        ".//esc:ServiceInstance",
        namespaces=NAMESPACES,
    )
    assert len(service_instance_nodes) == 1

    eherkenning_instance_node = service_instance_nodes[0]
    # Service instance eHerkenning
    service_id_node = eherkenning_instance_node.find(
        ".//esc:ServiceID",
        namespaces=NAMESPACES,
    )
    assert service_id_node.text == "urn:etoegang:DV:00000001112223330000:services:9050"

    service_url_node = eherkenning_instance_node.find(
        ".//esc:ServiceURL",
        namespaces=NAMESPACES,
    )
    assert service_url_node.text == "http://test-entity.id"

    privacy_url_node = eherkenning_instance_node.find(
        ".//esc:PrivacyPolicyURL",
        namespaces=NAMESPACES,
    )
    assert privacy_url_node.text == "http://test-privacy.nl"

    makelaar_id_node = eherkenning_instance_node.find(
        ".//esc:HerkenningsmakelaarId",
        namespaces=NAMESPACES,
    )
    assert makelaar_id_node.text == "00000003332221110000"

    key_name_node = eherkenning_instance_node.find(
        ".//ds:KeyName",
        namespaces=NAMESPACES,
    )
    assert key_name_node is not None
    certificate_node = eherkenning_instance_node.find(
        ".//ds:X509Certificate",
        namespaces=NAMESPACES,
    )
    assert certificate_node is not None

    classifier_node = eherkenning_instance_node.findall(
        ".//esc:Classifier",
        namespaces=NAMESPACES,
    )
    assert len(classifier_node) == 0



import binascii
from base64 import b64encode
from io import BytesIO
from typing import Union
from uuid import uuid4

from django.urls import reverse
from django.utils import timezone

from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_pem_x509_certificate
from furl.furl import furl
from lxml.builder import ElementMaker
from lxml.etree import Element, tostring
from onelogin.saml2.settings import OneLogin_Saml2_Settings

from ..models import EherkenningConfiguration
from ..settings import EHERKENNING_DS_XSD
from ..types import EHerkenningConfig, EHerkenningSAMLConfig, ServiceConfig
from ..utils import validate_xml
from .base import BaseSaml2Client, get_service_description, get_service_name

namespaces = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
    "esc": "urn:etoegang:1.13:service-catalog",
    "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
}
ESC = ElementMaker(namespace=namespaces["esc"], nsmap=namespaces)
DS = ElementMaker(namespace=namespaces["ds"], nsmap=namespaces)
SAML = ElementMaker(namespace=namespaces["saml"], nsmap=namespaces)
MD = ElementMaker(namespace=namespaces["md"], nsmap=namespaces)

xml_nl_lang = {"{http://www.w3.org/XML/1998/namespace}lang": "nl"}


def generate_dienst_catalogus_metadata(eherkenning_config=None):
    eherkenning_config = eherkenning_config or EherkenningConfiguration.get_solo()
    settings: EHerkenningConfig = eherkenning_config.as_dict()
    # ensure that the single language strings are output in both nl and en
    for service in settings["services"]:
        name = service["service_name"]
        service["service_name"] = {"nl": name, "en": name}

        description = service["service_description"]
        service["service_description"] = {"nl": description, "en": description}

        privacy_policy_url = service["privacy_policy_url"]
        service["privacy_policy_url"] = {"nl": privacy_policy_url}

        service_description_url = service["service_description_url"]
        service["service_description_url"] = (
            {"nl": service_description_url} if service_description_url else None
        )

    return create_service_catalogus(settings)


def generate_eherkenning_metadata():
    client = eHerkenningClient()
    client.saml2_setting_kwargs = {"sp_validation_only": True}
    metadata = client.create_metadata()
    return (
        b'<?xml version="1.0" encoding="UTF-8"?>' + metadata
        if not metadata.startswith(b"<?xml")
        else metadata
    )


def xml_datetime(d):
    return d.isoformat(timespec="seconds")


def create_language_elements(
    element_name: str,
    option_value: Union[dict[str, str], str, None],
    languages: Union[list[str], None] = None,
) -> list[Element]:
    """
    Convert a configuration option into zero or more eHerkenning dienstcatalogus
    elements

    :param element_name: Name of the XML element to be generated
    :param option_value: Configuration option being either a string or a dictionary
                        containing the language code as key, and the option as value.
    :param languages: A node for each language will be made if option_value is a string
    :return list: of etree elements
    """
    default_language = "en"

    if option_value is None:
        return []

    if isinstance(option_value, str):
        option_in_different_langs = (
            {language: option_value for language in languages}
            if languages
            else {default_language: option_value}
        )
    else:
        option_in_different_langs = option_value

    elements = []
    for lang, option in option_in_different_langs.items():
        xml_lang = {"{http://www.w3.org/XML/1998/namespace}lang": lang}
        elements.append(
            ESC(element_name, option, **xml_lang),
        )
    return elements


def create_service_catalogue(id, issue_instant, signature, service_provider):
    ns = namespaces["esc"]
    args = [
        signature,
        service_provider,
    ]
    kwargs = {
        "ID": id,
        f"{{{ns}}}IssueInstant": xml_datetime(issue_instant),
        f"{{{ns}}}Version": "urn:etoegang:1.13:53",
    }
    return ESC("ServiceCatalogue", *args, **kwargs)


def create_signature(id):
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Digital+signature
    """
    transforms = [
        DS(
            "Transform",
            Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature",
        ),
        DS("Transform", Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"),
    ]

    args = [
        DS(
            "SignedInfo",
            DS(
                "CanonicalizationMethod",
                Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#",
            ),
            DS(
                "SignatureMethod",
                Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            ),
            DS(
                "Reference",
                DS("Transforms", *transforms),
                DS("DigestMethod", Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DS("DigestValue"),
                URI=f"#{id}",
            ),
        ),
        DS("SignatureValue"),
    ]
    return DS("Signature", *args)


def create_service_provider(
    service_provider_id: str,
    organization_display_name: str,
    service_definitions: list,
    service_instances: list,
) -> Element:
    ns = namespaces["esc"]
    org_name_elements = create_language_elements(
        "OrganizationDisplayName", organization_display_name
    )

    args = [
        ESC("ServiceProviderID", service_provider_id),
        *org_name_elements,
        *service_definitions,
        *service_instances,
    ]
    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceProvider", *args, **kwargs)


def create_service_definition(
    service_uuid,
    service_name,
    service_description,
    service_description_url,
    loa,
    entity_concerned_types_allowed,
    requested_attributes,
    makelaar_oin,
    service_restrictions_allowed,
):
    service_name_elements = create_language_elements("ServiceName", service_name)
    service_description_elements = create_language_elements(
        "ServiceDescription", service_description
    )
    service_description_url_elements = create_language_elements(
        "ServiceDescriptionURL", service_description_url
    )

    ns = namespaces["esc"]
    args = [
        ESC("ServiceUUID", service_uuid),
        *service_name_elements,
        *service_description_elements,
        *service_description_url_elements,
        SAML("AuthnContextClassRef", loa),
        ESC("HerkenningsmakelaarId", makelaar_oin),
    ]

    for entity in entity_concerned_types_allowed:
        assert isinstance(entity, dict)

        kwargs = {}
        set_number = entity.get("set_number")
        if set_number:
            kwargs["setNumber"] = set_number
        args.append(
            ESC("EntityConcernedTypesAllowed", entity["name"], **kwargs),
        )

    if service_restrictions_allowed:
        args.append(ESC("ServiceRestrictionsAllowed", service_restrictions_allowed))

    for requested_attribute in requested_attributes:
        if isinstance(requested_attribute, dict):
            ra_kwargs = {}
            if "required" in requested_attribute:
                ra_kwargs["isRequired"] = (
                    "true" if requested_attribute["required"] else "false"
                )

            ra_args = []
            if not "purpose_statements" in requested_attribute:
                ra_args += create_language_elements("PurposeStatement", service_name)
            else:
                ra_args += create_language_elements(
                    "PurposeStatement", requested_attribute["purpose_statements"]
                )

            ra_kwargs["Name"] = requested_attribute["name"]
            args.append(
                ESC("RequestedAttribute", *ra_args, **ra_kwargs),
            )
        else:
            args.append(
                ESC("RequestedAttribute", Name=requested_attribute),
            )

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceDefinition", *args, **kwargs)


def create_service_instance(
    service_id,
    service_uuid,
    instance_of_service,
    service_url,
    privacy_policy_url,
    herkenningsmakelaars_id,
    classifiers,
    key_descriptor,
):
    ns = namespaces["esc"]

    service_url_elements = create_language_elements(
        "ServiceURL", service_url, languages=["nl", "en"]
    )
    privacy_url_elements = create_language_elements(
        "PrivacyPolicyURL", privacy_policy_url, languages=["nl", "en"]
    )

    args = [
        ESC("ServiceID", service_id),
        ESC("ServiceUUID", service_uuid),
        ESC("InstanceOfService", instance_of_service),
        *service_url_elements,
        *privacy_url_elements,
        ESC("HerkenningsmakelaarId", herkenningsmakelaars_id),
        ESC("SSOSupport", "false"),
        ESC("ServiceCertificate", key_descriptor),
    ]
    if len(classifiers) > 0:
        args.append(create_classifiers_element(classifiers))

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceInstance", *args, **kwargs)


def create_classifiers_element(classifiers: list) -> ElementMaker:
    classifiers_elements = []
    for classifier in classifiers:
        classifiers_elements.append(ESC("Classifier", classifier))
    return ESC("Classifiers", *classifiers_elements)


def create_key_descriptor(x509_certificate_content: bytes) -> ElementMaker:
    certificate = load_pem_x509_certificate(x509_certificate_content)
    key_name = binascii.hexlify(
        certificate.fingerprint(certificate.signature_hash_algorithm)
    ).decode("ascii")

    # grab the actual base64 data describding the certificate, but without the
    # BEGIN/END CERTIFICATE headers and footers and stripped of line breaks.
    certificate_content = certificate.public_bytes(serialization.Encoding.DER)
    key_descriptor_cert = b64encode(certificate_content).decode("ascii")

    args = [
        DS(
            "KeyInfo",
            DS("KeyName", key_name),
            DS("X509Data", DS("X509Certificate", key_descriptor_cert)),
        )
    ]
    kwargs = {"use": "encryption"}
    return MD("KeyDescriptor", *args, **kwargs)


def create_service_catalogus(conf: EHerkenningConfig, validate: bool = True) -> bytes:
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Service+catalog
    """
    with conf["cert_file"].open("rb") as cert_file:
        x509_certificate_content: bytes = cert_file.read()

    sc_id = str(uuid4())
    service_provider_id = conf["oin"]
    organization_display_name = conf["organization_name"]

    service_definitions = []
    service_instances = []
    for service in conf["services"]:
        key_descriptor = create_key_descriptor(x509_certificate_content)

        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceUUID
        service_uuid = service["service_uuid"]
        service_name = service["service_name"]
        service_description = service["service_description"]
        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceID
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_instance_uuid = service["service_instance_uuid"]

        service_url = service.get(
            "service_url",
        )
        privacy_policy_url = service.get(
            "privacy_policy_url",
        )
        service_description_url = service.get(
            "service_description_url",
        )
        herkenningsmakelaars_id = service.get(
            "herkenningsmakelaars_id",
        )
        entity_concerned_types_allowed = service.get("entity_concerned_types_allowed")
        service_restrictions_allowed = service.get("service_restrictions_allowed")
        requested_attributes = service.get("requested_attributes", [])
        classifiers = service.get("classifiers", [])

        service_definition = create_service_definition(
            service_uuid,
            service_name,
            service_description,
            service_description_url,
            # https://afsprakenstelsel.etoegang.nl/display/as/Level+of+assurance
            service["loa"],
            entity_concerned_types_allowed,
            requested_attributes,
            herkenningsmakelaars_id,
            service_restrictions_allowed,
        )
        service_instance = create_service_instance(
            service_id,
            service_instance_uuid,
            service_uuid,
            service_url,
            privacy_policy_url,
            herkenningsmakelaars_id,
            classifiers,
            key_descriptor,
        )

        service_definitions.append(service_definition)
        service_instances.append(service_instance)

    signature = create_signature(sc_id)
    service_provider = create_service_provider(
        service_provider_id,
        organization_display_name,
        service_definitions,
        service_instances,
    )
    xml = create_service_catalogue(sc_id, timezone.now(), signature, service_provider)

    catalogus = tostring(xml, pretty_print=True, xml_declaration=True, encoding="utf-8")
    if validate:
        errors = validate_xml(BytesIO(catalogus), EHERKENNING_DS_XSD)
        assert errors is None, errors
    return catalogus


def get_metadata_eherkenning_requested_attributes(
    conf: ServiceConfig, service_id: str
) -> list[dict]:
    # There needs to be a RequestedAttribute element where the name is the ServiceID
    # https://afsprakenstelsel.etoegang.nl/display/as/DV+metadata+for+HM
    requested_attributes = [{"name": service_id, "isRequired": False}]
    for requested_attribute in conf.get("requested_attributes", []):
        if isinstance(requested_attribute, dict):
            requested_attributes.append(
                {
                    "name": requested_attribute["name"],
                    "isRequired": requested_attribute["required"],
                }
            )
        else:
            requested_attributes.append(
                {
                    "name": requested_attribute,
                    "isRequired": True,
                }
            )

    return requested_attributes


def create_attribute_consuming_services(conf: EHerkenningConfig) -> list[dict]:
    attribute_consuming_services = []

    for service in conf["services"]:
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_name = get_service_name(service)
        service_description = get_service_description(service)
        requested_attributes = get_metadata_eherkenning_requested_attributes(
            service, service_id
        )

        attribute_consuming_services.append(
            {
                "index": service["attribute_consuming_service_index"],
                "serviceName": service_name,
                "serviceDescription": service_description,
                "requestedAttributes": requested_attributes,
                "language": service.get("language", "nl"),
            }
        )
    return attribute_consuming_services


class eHerkenningClient(BaseSaml2Client):
    cache_key_prefix = "eherkenning"
    cache_timeout = 60 * 60  # 1 hour

    @property
    def conf(self) -> EHerkenningConfig:
        if not hasattr(self, "_conf"):
            db_config = EherkenningConfiguration.get_solo()
            self._conf = db_config.as_dict()
            self._conf.setdefault("acs_path", reverse("eherkenning:acs"))
        return self._conf

    def create_config_dict(self, conf: EHerkenningConfig) -> EHerkenningSAMLConfig:
        config_dict: EHerkenningSAMLConfig = super().create_config_dict(conf)

        attribute_consuming_services = create_attribute_consuming_services(conf)
        with conf["cert_file"].open("r") as cert_file, conf["key_file"].open(
            "r"
        ) as key_file:
            certificate = cert_file.read()
            privkey = key_file.read()
        acs_url = furl(conf["base_url"]) / conf["acs_path"]
        config_dict.update(
            {
                "sp": {
                    # Identifier of the SP entity  (must be a URI)
                    "entityId": conf["entity_id"],
                    # Specifies info about where and how the <AuthnResponse> message MUST be
                    # returned to the requester, in this case our SP.
                    "assertionConsumerService": {
                        "url": acs_url.url,
                        "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                    },
                    "attributeConsumingServices": attribute_consuming_services,
                    "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                    "x509cert": certificate,
                    "privateKey": privkey,
                    "privateKeyPassphrase": conf.get("key_passphrase", None),
                },
            }
        )
        return config_dict

    def create_config(
        self, config_dict: EHerkenningSAMLConfig
    ) -> OneLogin_Saml2_Settings:
        config_dict["security"].update(
            {
                # See comment in the python3-saml for in  OneLogin_Saml2_Response.validate_num_assertions (onelogin/saml2/response.py)
                # for why we need this option.
                "disableSignatureWrappingProtection": True,
                # For eHerkenning, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContext": False,
            }
        )
        return super().create_config(config_dict)

    def create_authn_request(
        self, request, return_to=None, attr_consuming_service_index=None, **kwargs
    ):
        return super().create_authn_request(
            request,
            return_to=return_to,
            force_authn=True,
            is_passive=False,
            set_nameid_policy=False,
            name_id_value_req=None,
            attr_consuming_service_index=attr_consuming_service_index,
        )



from django.core.management import BaseCommand

from digid_eherkenning.models.digid import DigidConfiguration
from digid_eherkenning.models.eherkenning import EherkenningConfiguration


class Command(BaseCommand):
    help = "Updates the stored metadata file and prepopulates the db fields."

    def add_arguments(self, parser):
        parser.add_argument(
            "config_model",
            type=str,
            choices=["digid", "eherkenning"],
            help="Update the DigiD or Eherkenning configuration metadata.",
        )

    def handle(self, **options):
        if options["config_model"] == "digid":
            config = DigidConfiguration.get_solo()
        elif options["config_model"] == "eherkenning":
            config = EherkenningConfiguration.get_solo()

        if config.metadata_file_source:
            config.save()
            self.stdout.write(self.style.SUCCESS("Update was successful"))
        else:
            self.stdout.write(
                self.style.WARNING("Update failed, no metadata file source found")
            )



from django.views import View


def get_next_page(view: View) -> str:
    try:
        return view.get_default_redirect_url()  # type: ignore
    except AttributeError:
        return view.get_next_page()  # type: ignore



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from privates.admin import PrivateMediaMixin
from privates.widgets import PrivateFileWidget
from solo.admin import SingletonModelAdmin

from .models import DigidConfiguration, EherkenningConfiguration


class CustomPrivateFileWidget(PrivateFileWidget):
    template_name = "admin/digid_eherkenning/widgets/custom_file_input.html"


class CustomPrivateMediaMixin(PrivateMediaMixin):
    private_media_file_widget = CustomPrivateFileWidget


@admin.register(DigidConfiguration)
class DigidConfigurationAdmin(CustomPrivateMediaMixin, SingletonModelAdmin):
    readonly_fields = ("idp_service_entity_id",)
    fieldsets = (
        (
            _("X.509 Certificate"),
            {
                "fields": (
                    "certificate",
                    "key_passphrase",
                ),
            },
        ),
        (
            _("Identity provider"),
            {
                "fields": (
                    "metadata_file_source",
                    "idp_service_entity_id",
                    "idp_metadata_file",
                ),
            },
        ),
        (
            _("SAML configuration"),
            {
                "fields": (
                    "entity_id",
                    "base_url",
                    "artifact_resolve_content_type",
                    "want_assertions_signed",
                    "want_assertions_encrypted",
                    "signature_algorithm",
                    "digest_algorithm",
                ),
            },
        ),
        (
            _("Service details"),
            {
                "fields": (
                    "service_name",
                    "service_description",
                    "requested_attributes",
                    "attribute_consuming_service_index",
                    "slo",
                ),
            },
        ),
        (
            _("Organization details"),
            {
                "fields": (
                    "technical_contact_person_telephone",
                    "technical_contact_person_email",
                    "organization_url",
                    "organization_name",
                ),
            },
        ),
    )
    change_form_template = "admin/digid_eherkenning/digidconfiguration/change_form.html"
    private_media_fields = ("idp_metadata_file",)


@admin.register(EherkenningConfiguration)
class EherkenningConfigurationAdmin(CustomPrivateMediaMixin, SingletonModelAdmin):
    readonly_fields = ("idp_service_entity_id",)
    fieldsets = (
        (
            _("X.509 Certificate"),
            {
                "fields": (
                    "certificate",
                    "key_passphrase",
                ),
            },
        ),
        (
            _("Identity provider"),
            {
                "fields": (
                    "metadata_file_source",
                    "idp_service_entity_id",
                    "idp_metadata_file",
                ),
            },
        ),
        (
            _("SAML configuration"),
            {
                "fields": (
                    "entity_id",
                    "base_url",
                    "artifact_resolve_content_type",
                    "want_assertions_signed",
                    "want_assertions_encrypted",
                    "signature_algorithm",
                    "digest_algorithm",
                ),
            },
        ),
        (
            _("Service details"),
            {
                "fields": (
                    "service_name",
                    "service_description",
                    "oin",
                    "makelaar_id",
                    "privacy_policy",
                    "service_language",
                ),
            },
        ),
        (
            _("eHerkenning"),
            {
                "fields": (
                    "eh_requested_attributes",
                    "eh_attribute_consuming_service_index",
                    "eh_service_uuid",
                    "eh_service_instance_uuid",
                    "eh_loa",
                ),
            },
        ),
        (
            _("eIDAS"),
            {
                "fields": (
                    "no_eidas",
                    "eidas_requested_attributes",
                    "eidas_attribute_consuming_service_index",
                    "eidas_service_uuid",
                    "eidas_service_instance_uuid",
                    "eidas_loa",
                ),
            },
        ),
        (
            _("Organization details"),
            {
                "fields": (
                    "technical_contact_person_telephone",
                    "technical_contact_person_email",
                    "organization_url",
                    "organization_name",
                ),
            },
        ),
    )
    change_form_template = (
        "admin/digid_eherkenning/eherkenningconfiguration/change_form.html"
    )
    private_media_fields = ("idp_metadata_file",)



# Generated by Django 4.2.10 on 2024-03-08 08:45

from django.db import migrations, models

import digid_eherkenning.choices


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning",
            "0007_eherkenningconfiguration_service_description_url",
        ),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="eherkenningconfiguration",
            name="valid_loa",
        ),
        migrations.RenameField(
            model_name="eherkenningconfiguration",
            old_name="loa",
            new_name="eh_loa",
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="eh_loa",
            field=models.CharField(
                choices=[
                    ("urn:etoegang:core:assurance-class:loa1", "Non existent (1)"),
                    ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                    ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                    ("urn:etoegang:core:assurance-class:loa3", "Substantial (3)"),
                    ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                ],
                default="urn:etoegang:core:assurance-class:loa3",
                help_text="Level of Assurance (LoA) to use for the eHerkenning service.",
                max_length=100,
                verbose_name="eHerkenning LoA",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfiguration",
            name="eidas_loa",
            field=models.CharField(
                choices=[
                    ("urn:etoegang:core:assurance-class:loa1", "Non existent (1)"),
                    ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                    ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                    ("urn:etoegang:core:assurance-class:loa3", "Substantial (3)"),
                    ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                ],
                default="urn:etoegang:core:assurance-class:loa3",
                help_text="Level of Assurance (LoA) to use for the eIDAS service.",
                max_length=100,
                verbose_name="eIDAS LoA",
            ),
        ),
        migrations.AddConstraint(
            model_name="eherkenningconfiguration",
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ("eh_loa__in", digid_eherkenning.choices.AssuranceLevels),
                        ("eidas_loa__in", digid_eherkenning.choices.AssuranceLevels),
                    )
                ),
                name="valid_loa",
            ),
        ),
    ]



"""
XML parsing with DTD/Entities blocking.

Inspired by https://github.com/mvantellingen/python-zeep/pull/1179/ as their solution
for the deprecated defusedxml.lxml module and the defaults applied in defusedxml.lxml.
"""

from lxml.etree import XMLParser, parse as _parse


def parse(source):
    """
    Parse an LXML etree from source without resolving entities.

    Resolving entities is a security risk, which is why we disable it.
    """
    parser = XMLParser(resolve_entities=False)
    return _parse(source, parser)



# Generated by Django 4.2.10 on 2024-03-04 13:08

from django.db import migrations, models
import digid_eherkenning.models.eherkenning


class Migration(migrations.Migration):

    dependencies = [
        ("digid_eherkenning", "0006_digidconfiguration_metadata_file_source_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="eherkenningconfiguration",
            name="service_description_url",
            field=models.URLField(
                default="",
                help_text="The URL where the service description can be found.",
                max_length=255,
                verbose_name="service description URL",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="eh_requested_attributes",
            field=models.JSONField(
                default=digid_eherkenning.models.eherkenning.get_default_requested_attributes_eherkenning,
                help_text="A list of additional requested attributes. A single requested attribute can be a string (the name of the attribute) or an object with keys 'name' and 'required', where 'name' is a string and 'required' a boolean'.",
                verbose_name="requested attributes",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="eidas_requested_attributes",
            field=models.JSONField(
                default=digid_eherkenning.models.eherkenning.get_default_requested_attributes_eidas,
                help_text="A list of additional requested attributes. A single requested attribute can be a string (the name of the attribute) or an object with keys 'name' and 'required', where 'name' is a string and 'required' a boolean'.",
                verbose_name="requested attributes",
            ),
        ),
    ]



# Generated by Django 4.2.4 on 2023-08-15 13:47

import digid_eherkenning.choices
from django.db import migrations, models
from ..choices import AssuranceLevels


def set_default_loa(apps, schema_editor):
    EherkenningConfiguration = apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )
    EherkenningConfiguration.objects.filter(loa="").update(loa=AssuranceLevels.low_plus)


class Migration(migrations.Migration):
    dependencies = [
        ("digid_eherkenning", "0003_digidconfiguration_artifact_resolve_content_type"),
    ]

    operations = [
        migrations.RunPython(set_default_loa),
        migrations.AlterField(
            model_name="eherkenningconfiguration",
            name="loa",
            field=models.CharField(
                choices=[
                    ("urn:etoegang:core:assurance-class:loa1", "Non existent (1)"),
                    ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                    ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                    ("urn:etoegang:core:assurance-class:loa3", "Substantial (3)"),
                    ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                ],
                default="urn:etoegang:core:assurance-class:loa3",
                help_text="Level of Assurance (LoA) to use for all the services.",
                max_length=100,
                verbose_name="LoA",
            ),
        ),
        migrations.AddConstraint(
            model_name="eherkenningconfiguration",
            constraint=models.CheckConstraint(
                check=models.Q(("loa__in", digid_eherkenning.choices.AssuranceLevels)),
                name="valid_loa",
            ),
        ),
    ]



from io import StringIO

from django.core.management import CommandError, call_command
from django.test import TestCase

import pytest
from lxml import etree
from privates.test import temp_private_root

from digid_eherkenning.models import EherkenningConfiguration
from digid_eherkenning.saml2.eherkenning import (
    eHerkenningClient,
    generate_eherkenning_metadata,
)

from .conftest import EHERKENNING_TEST_CERTIFICATE_FILE, EHERKENNING_TEST_KEY_FILE
from .mixins import EherkenningMetadataMixin

NAME_SPACES = {
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
}


@temp_private_root()
class EHerkenningMetadataManagementCommandTests(TestCase):
    def test_generate_metadata_all_options_specified(self):
        stdout = StringIO()

        call_command(
            "generate_eherkenning_metadata",
            "--no-save-config",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(EHERKENNING_TEST_KEY_FILE),
            cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
            signature_algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            digest_algorithm="http://www.w3.org/2001/04/xmlenc#sha256",
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            eh_attribute_consuming_service_index="9050",
            eidas_attribute_consuming_service_index="9051",
            oin="00000001112223330000",
            service_name="Test Service Name",
            service_description="Test Service Description",
            technical_contact_person_telephone="06123123123",
            technical_contact_person_email="test@test.nl",
            organization_name="Test organisation",
            organization_url="http://test-organisation.nl",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output.encode("utf-8"))

        self.assertEqual(
            "http://test-entity.id", entity_descriptor_node.attrib["entityID"]
        )

        sspo_descriptor_node = entity_descriptor_node.find(
            ".//md:SPSSODescriptor",
            namespaces=NAME_SPACES,
        )

        self.assertEqual("true", sspo_descriptor_node.attrib["AuthnRequestsSigned"])
        self.assertEqual("true", sspo_descriptor_node.attrib["WantAssertionsSigned"])

        certificate_node = entity_descriptor_node.find(
            ".//ds:X509Certificate",
            namespaces=NAME_SPACES,
        )
        self.assertIn(
            "MIIC0DCCAbigAwIBAgIUEjGmfCGa1cOiTi+UKtDQVtySOHUwDQYJKoZIhvcNAQEL",
            certificate_node.text,
        )

        signature_algorithm_node = entity_descriptor_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = entity_descriptor_node.find(
            ".//ds:DigestMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        assertion_consuming_service_node = entity_descriptor_node.find(
            ".//md:AssertionConsumerService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://test-entity.id/eherkenning/acs/",
            assertion_consuming_service_node.attrib["Location"],
        )

        attribute_consuming_service_nodes = entity_descriptor_node.findall(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(2, len(attribute_consuming_service_nodes))

        eh_attribute_consuming_service_node = attribute_consuming_service_nodes[0]
        eidas_attribute_consuming_service_node = attribute_consuming_service_nodes[1]

        self.assertEqual(
            "urn:etoegang:DV:00000001112223330000:services:9050",
            eh_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "urn:etoegang:DV:00000001112223330000:services:9051",
            eidas_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name (eIDAS)",
            eidas_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eidas_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_name_node.text)

        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_display_node.text)

        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("http://test-organisation.nl", organisation_url_node.text)

        contact_person_node = entity_descriptor_node.find(
            ".//md:ContactPerson",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("technical", contact_person_node.attrib["contactType"])

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("test@test.nl", contact_email_node.text)

        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("06123123123", contact_telephone_node.text)

    def test_missing_required_properties(self):
        with self.assertRaises(CommandError):
            call_command("generate_eherkenning_metadata")

    def test_contact_telephone_no_email(self):
        stdout = StringIO()

        call_command(
            "generate_eherkenning_metadata",
            "--no-save-config",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(EHERKENNING_TEST_KEY_FILE),
            cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
            oin="00000001112223330000",
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            technical_contact_person_telephone="06123123123",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output.encode("utf-8"))

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(contact_email_node)
        self.assertIsNone(contact_telephone_node)

    def test_organisation_url_no_service(self):
        stdout = StringIO()

        call_command(
            "generate_eherkenning_metadata",
            "--no-save-config",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            oin="00000001112223330000",
            key_file=str(EHERKENNING_TEST_KEY_FILE),
            cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            organization_url="http://test-organisation.nl",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output.encode("utf-8"))

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(organisation_name_node)
        self.assertIsNone(organisation_display_node)
        self.assertIsNone(organisation_url_node)

    def test_no_eidas_service(self):
        stdout = StringIO()

        call_command(
            "generate_eherkenning_metadata",
            "--no-save-config",
            stdout=stdout,
            want_assertions_encrypted=True,
            want_assertions_signed=True,
            key_file=str(EHERKENNING_TEST_KEY_FILE),
            cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
            signature_algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            digest_algorithm="http://www.w3.org/2001/04/xmlenc#sha256",
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            eh_attribute_consuming_service_index="9050",
            oin="00000001112223330000",
            no_eidas=True,
            service_name="Test Service Name",
            service_description="Test Service Description",
            technical_contact_person_telephone="06123123123",
            technical_contact_person_email="test@test.nl",
            organization_name="Test organisation",
            organization_url="http://test-organisation.nl",
            test=True,
        )

        output = stdout.getvalue()
        entity_descriptor_node = etree.XML(output.encode("utf-8"))

        attribute_consuming_service_nodes = entity_descriptor_node.findall(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(1, len(attribute_consuming_service_nodes))

        eh_attribute_consuming_service_node = attribute_consuming_service_nodes[0]

        self.assertEqual(
            "urn:etoegang:DV:00000001112223330000:services:9050",
            eh_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )

    def test_management_command_and_update_config(self):
        stdout = StringIO()
        assert not EherkenningConfiguration.objects.exists()

        call_command(
            "generate_eherkenning_metadata",
            "--save-config",
            "--want-assertions-encrypted",
            "--no-only-assertions-signed",
            ["--eh-attribute-consuming-service-index", "1"],
            key_file=str(EHERKENNING_TEST_KEY_FILE),
            cert_file=str(EHERKENNING_TEST_CERTIFICATE_FILE),
            entity_id="http://test-entity.id",
            base_url="http://test-entity.id",
            service_name="Test Service Name",
            service_description="Test Service Description",
            oin="01234567890123456789",
            stdout=stdout,
            test=True,
        )

        self.assertTrue(EherkenningConfiguration.objects.exists())
        config = EherkenningConfiguration.get_solo()
        self.assertTrue(config.want_assertions_encrypted)
        self.assertFalse(config.want_assertions_signed)
        self.assertEqual(config.oin, "01234567890123456789")
        self.assertEqual(config.service_name, "Test Service Name")
        self.assertEqual(config.service_description, "Test Service Description")
        self.assertEqual(config.eh_attribute_consuming_service_index, "1")

        self.assertIsNotNone(config.certificate)

        with config.certificate.private_key.open("rb") as privkey:
            with EHERKENNING_TEST_KEY_FILE.open("rb") as source_privkey:
                self.assertEqual(privkey.read(), source_privkey.read())

        with config.certificate.public_certificate.open("rb") as cert:
            with EHERKENNING_TEST_CERTIFICATE_FILE.open("rb") as source_cert:
                self.assertEqual(cert.read(), source_cert.read())


@pytest.mark.usefixtures("eherkenning_config", "temp_private_root")
class EHerkenningClientTests(TestCase):
    def test_attribute_consuming_services_with_non_required_requested_attribute(self):
        config = EherkenningConfiguration.get_solo()
        config.eh_requested_attributes = [{"name": "Test Attribute", "required": False}]
        config.save()

        eherkenning_client = eHerkenningClient()
        metadata = eherkenning_client.create_metadata()

        tree = etree.XML(metadata)
        namespace = {
            "md": "urn:oasis:names:tc:SAML:2.0:metadata",
        }
        attribute_consuming_service_node = tree.find(
            ".//md:AttributeConsumingService",
            namespaces=namespace,
        )
        requested_attribute_nodes = attribute_consuming_service_node.findall(
            ".//md:RequestedAttribute",
            namespaces=namespace,
        )
        self.assertEqual(2, len(requested_attribute_nodes))

        default_requested_attribute_node = requested_attribute_nodes[0]
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000000:services:1",
            default_requested_attribute_node.attrib["Name"],
        )
        self.assertNotIn("isRequired", default_requested_attribute_node.attrib)

        requested_attribute_node = requested_attribute_nodes[1]
        self.assertEqual("Test Attribute", requested_attribute_node.attrib["Name"])
        self.assertNotIn("isRequired", requested_attribute_node.attrib)

    def test_attribute_consuming_services_with_required_requested_attribute(self):
        config = EherkenningConfiguration.get_solo()
        config.eh_requested_attributes = [{"name": "Test Attribute", "required": True}]
        config.save()

        eherkenning_client = eHerkenningClient()
        metadata = eherkenning_client.create_metadata()

        tree = etree.XML(metadata)
        namespace = {
            "md": "urn:oasis:names:tc:SAML:2.0:metadata",
            "xml": "http://www.w3.org/XML/1998/namespace",
        }
        attribute_consuming_service_node = tree.find(
            ".//md:AttributeConsumingService",
            namespaces=namespace,
        )
        requested_attribute_nodes = attribute_consuming_service_node.findall(
            ".//md:RequestedAttribute",
            namespaces=namespace,
        )
        self.assertEqual(2, len(requested_attribute_nodes))

        default_requested_attribute_node = requested_attribute_nodes[0]
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000000:services:1",
            default_requested_attribute_node.attrib["Name"],
        )
        self.assertNotIn("isRequired", default_requested_attribute_node.attrib)

        requested_attribute_node = requested_attribute_nodes[1]
        self.assertEqual("Test Attribute", requested_attribute_node.attrib["Name"])
        self.assertEqual("true", requested_attribute_node.attrib["isRequired"])

    def test_attribute_consuming_services_dutch(self):
        config = EherkenningConfiguration.get_solo()
        config.no_eidas = True
        config.service_language = "en"
        config.save()

        eherkenning_client = eHerkenningClient()
        metadata = eherkenning_client.create_metadata()

        tree = etree.XML(metadata)
        namespace = {
            "md": "urn:oasis:names:tc:SAML:2.0:metadata",
        }
        attribute_consuming_service_node = tree.find(
            ".//md:AttributeConsumingService",
            namespaces=namespace,
        )
        service_name_node = attribute_consuming_service_node.find(
            ".//md:ServiceName",
            namespaces=namespace,
        )
        service_description_node = attribute_consuming_service_node.find(
            ".//md:ServiceDescription",
            namespaces=namespace,
        )
        requested_attribute_node = attribute_consuming_service_node.find(
            ".//md:RequestedAttribute",
            namespaces=namespace,
        )

        self.assertEqual(
            "urn:etoegang:DV:00000000000000000000:services:1",
            requested_attribute_node.attrib["Name"],
        )
        self.assertEqual(
            "en", service_name_node.attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        )
        self.assertEqual(
            "en",
            service_description_node.attrib[
                "{http://www.w3.org/XML/1998/namespace}lang"
            ],
        )

    def test_with_bogus_or_bad_idp_metadata(self):
        config = EherkenningConfiguration.get_solo()
        # different idp_service_entity_id than what's in the metadata
        config.idp_service_entity_id = (
            "not:urn:etoegang:HM:00000003520354760000:entities:9632"
        )
        config.save()

        try:
            generate_eherkenning_metadata()
        except Exception:
            self.fail("Metadata generation should not have crashed")


@pytest.mark.usefixtures("eherkenning_config_defaults", "temp_private_root")
class EHerkenningMetadataTests(EherkenningMetadataMixin, TestCase):
    def test_generate_metadata_all_options_specified(self):
        self.eherkenning_config.signature_algorithm = (
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
        )
        self.eherkenning_config.digest_algorithm = (
            "http://www.w3.org/2001/04/xmlenc#sha256"
        )
        self.eherkenning_config.technical_contact_person_telephone = "06123123123"
        self.eherkenning_config.technical_contact_person_email = "test@test.nl"
        self.eherkenning_config.organization_name = "Test organisation"
        self.eherkenning_config.organization_url = "http://test-organisation.nl"
        self.eherkenning_config.save()

        eherkenning_metadata = generate_eherkenning_metadata()
        self.assertEqual(eherkenning_metadata[:5], b"<?xml")
        entity_descriptor_node = etree.XML(eherkenning_metadata)

        self.assertEqual(
            "http://test-entity.id", entity_descriptor_node.attrib["entityID"]
        )

        sspo_descriptor_node = entity_descriptor_node.find(
            ".//md:SPSSODescriptor",
            namespaces=NAME_SPACES,
        )

        self.assertEqual("true", sspo_descriptor_node.attrib["AuthnRequestsSigned"])
        self.assertEqual("true", sspo_descriptor_node.attrib["WantAssertionsSigned"])

        certificate_node = entity_descriptor_node.find(
            ".//ds:X509Certificate",
            namespaces=NAME_SPACES,
        )
        self.assertIn(
            "MIIC0DCCAbigAwIBAgIUEjGmfCGa1cOiTi+UKtDQVtySOHUwDQYJKoZIhvcNAQEL",
            certificate_node.text,
        )

        signature_algorithm_node = entity_descriptor_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = entity_descriptor_node.find(
            ".//ds:DigestMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        assertion_consuming_service_node = entity_descriptor_node.find(
            ".//md:AssertionConsumerService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://test-entity.id/eherkenning/acs/",
            assertion_consuming_service_node.attrib["Location"],
        )

        attribute_consuming_service_nodes = entity_descriptor_node.findall(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(2, len(attribute_consuming_service_nodes))

        eh_attribute_consuming_service_node = attribute_consuming_service_nodes[0]
        eidas_attribute_consuming_service_node = attribute_consuming_service_nodes[1]

        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9050",
            eh_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9051",
            eidas_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name (eIDAS)",
            eidas_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eidas_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_name_node.text)

        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_display_node.text)

        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("http://test-organisation.nl", organisation_url_node.text)

        contact_person_node = entity_descriptor_node.find(
            ".//md:ContactPerson",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("technical", contact_person_node.attrib["contactType"])

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("test@test.nl", contact_email_node.text)

        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("06123123123", contact_telephone_node.text)

    def test_contact_telephone_no_email(self):
        self.eherkenning_config.technical_contact_person_telephone = "06123123123"
        self.eherkenning_config.save()

        eherkenning_metadata = generate_eherkenning_metadata()
        entity_descriptor_node = etree.XML(eherkenning_metadata)

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(contact_email_node)
        self.assertIsNone(contact_telephone_node)

    def test_organisation_url_no_service(self):
        self.eherkenning_config.organization_url = "http://test-organisation.nl"
        self.eherkenning_config.save()

        eherkenning_metadata = generate_eherkenning_metadata()
        entity_descriptor_node = etree.XML(eherkenning_metadata)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(organisation_name_node)
        self.assertIsNone(organisation_display_node)
        self.assertIsNone(organisation_url_node)

    def test_no_eidas_service(self):
        self.eherkenning_config.no_eidas = True
        self.eherkenning_config.save()

        eherkenning_metadata = generate_eherkenning_metadata()

        entity_descriptor_node = etree.XML(eherkenning_metadata)

        attribute_consuming_service_nodes = entity_descriptor_node.findall(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(1, len(attribute_consuming_service_nodes))

        eh_attribute_consuming_service_node = attribute_consuming_service_nodes[0]

        self.assertEqual(
            "urn:etoegang:DV:00000000000000000011:services:9050",
            eh_attribute_consuming_service_node.find(
                ".//md:RequestedAttribute", namespaces=NAME_SPACES
            ).attrib["Name"],
        )
        self.assertEqual(
            "Test Service Name",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceName", namespaces=NAME_SPACES
            ).text,
        )
        self.assertEqual(
            "Test Service Description",
            eh_attribute_consuming_service_node.find(
                ".//md:ServiceDescription", namespaces=NAME_SPACES
            ).text,
        )



from unittest.mock import patch

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils.translation import gettext as _

import pytest

from digid_eherkenning.models import DigidConfiguration, EherkenningConfiguration

from .conftest import (
    DIGID_TEST_METADATA_FILE_SLO_POST,
    DIGID_TEST_METADATA_FILE_SLO_POST_2,
    EHERKENNING_TEST_METADATA_FILE,
)


@pytest.mark.usefixtures("digid_config_defaults", "temp_private_root")
class BaseModelTests(TestCase):
    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_fields_are_populated_on_digid_save(self, get_matadata):
        config = DigidConfiguration.get_solo()

        with DIGID_TEST_METADATA_FILE_SLO_POST.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_matadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://was-preprod1.digid.nl/saml/idp/metadata/with-slo"
            )
            config.save()

        self.assertTrue(get_matadata.called_once())
        self.assertEqual(
            config.idp_metadata_file.read().decode("utf-8"), metadata_content
        )
        self.assertEqual(
            config.idp_service_entity_id,
            "https://was-preprod1.digid.nl/saml/idp/metadata/with-slo",
        )

    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_fields_are_populated_on_eherkennig_save(self, get_matadata):
        config = EherkenningConfiguration.get_solo()

        with EHERKENNING_TEST_METADATA_FILE.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_matadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://eh01.staging.iwelcome.nl/broker/sso/1.13"
            )
            config.save()

        self.assertTrue(get_matadata.called_once())
        self.assertEqual(
            config.idp_metadata_file.read().decode("utf-8"), metadata_content
        )
        self.assertEqual(
            config.idp_service_entity_id,
            "urn:etoegang:HM:00000003520354760000:entities:9632",
        )

    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_no_fetching_xml_when_no_file_source_change(self, get_matadata):
        config = DigidConfiguration.get_solo()

        with DIGID_TEST_METADATA_FILE_SLO_POST.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_matadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://was-preprod1.digid.nl/saml/idp/metadata"
            )
            config.save()

        config.organization_name = "test"
        config.save()

        # Make sure we don't try to fetch and parse again the xml file, since there is no update
        self.assertTrue(get_matadata.called_once())

    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_wrong_xml_format_raises_validation_error(self, get_matadata):
        config = DigidConfiguration.get_solo()

        with DIGID_TEST_METADATA_FILE_SLO_POST.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_matadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://was-preprod1.digid.nl/saml/idp/metadata"
            )
            config.save()

        get_matadata.return_value = "wrong xml format"
        config.metadata_file_source = "https://example.com"

        with self.assertRaisesMessage(
            ValidationError,
            _(
                "Failed to parse the metadata, got error: Start tag expected, '<' not found, line 1, column 1"
            ),
        ):
            config.save()

    @patch("onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.parse")
    @patch(
        "onelogin.saml2.idp_metadata_parser.OneLogin_Saml2_IdPMetadataParser.get_metadata"
    )
    def test_no_idp_in_xml_raises_validation_error(self, get_matadata, parse):
        config = DigidConfiguration.get_solo()

        with DIGID_TEST_METADATA_FILE_SLO_POST_2.open("rb") as metadata_file:
            metadata_content = metadata_file.read().decode("utf-8")
            get_matadata.return_value = metadata_content
            config.metadata_file_source = (
                "https://was-preprod1.digid.nl/saml/idp/metadata"
            )
            parse.return_value = {"test_no_idp": ""}

        with self.assertRaisesMessage(
            ValidationError,
            _(
                "Could not find any identity provider information in the metadata at the provided URL."
            ),
        ):
            config.save()



import logging
from urllib.parse import urlencode

from django.http import HttpResponseBadRequest
from django.urls import reverse
from django.views.generic import FormView, TemplateView

from furl import furl

from digid_eherkenning.mock import conf
from digid_eherkenning.mock.idp.forms import PasswordLoginForm

logger = logging.getLogger(__name__)


class _BaseIDPViewMixin(TemplateView):
    page_title = "DigiD: Inloggen"

    def dispatch(self, request, *args, **kwargs):
        # we pass these variables through the URL instead of dealing with POST and sessions
        self.acs_url = self.request.GET.get("acs")
        self.next_url = self.request.GET.get("next")
        self.cancel_url = self.request.GET.get("cancel") or self.request.GET.get(
            "next", ""
        )

        if not self.acs_url:
            logger.debug("missing 'acs' parameter")
            return HttpResponseBadRequest("missing 'acs' parameter")
        if not self.next_url:
            logger.debug("missing 'next' parameter")
            return HttpResponseBadRequest("missing 'next' parameter")
        if not self.cancel_url:
            logger.debug("missing 'cancel' parameter")
            return HttpResponseBadRequest("missing 'cancel' parameter")

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        return {
            "app_title": conf.get_app_title(),
            "page_title": self.page_title,
            **super().get_context_data(**kwargs),
        }


class DigiDMockIDPLoginView(_BaseIDPViewMixin):
    """
    Login method choices pages
    """

    template_name = "digid_eherkenning/mock/login.html"
    page_title = "DigiD: Inloggen | Keuze"

    def get_context_data(self, **kwargs):
        params = {
            "acs": self.acs_url,
            "next": self.next_url,
            "cancel": self.cancel_url,
        }
        return {
            "cancel_url": params["cancel"],
            "password_login_url": f"{reverse('digid-mock:password')}?{urlencode(params)}",
            **super().get_context_data(**kwargs),
        }


class DigiDMockIDPPasswordLoginView(_BaseIDPViewMixin, FormView):
    """
    Username/password login page
    """

    template_name = "digid_eherkenning/mock/password.html"
    page_title = "DigiD: Inloggen | Gebruikersnaam en wachtwoord"

    form_class = PasswordLoginForm

    def form_valid(self, form):
        self.bsn = form.cleaned_data["auth_name"]
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        params = {
            "acs": self.acs_url,
            "next": self.next_url,
            "cancel": self.cancel_url,
        }
        return {
            "action_url": f"{reverse('digid-mock:password')}?{urlencode(params)}",
            "back_url": f"{reverse('digid-mock:login')}?{urlencode(params)}",
            **super().get_context_data(**kwargs),
        }

    def get_success_url(self):
        params = {
            "next": self.next_url,
            "bsn": str(self.bsn),
        }
        success_url = furl(self.acs_url)
        success_url.args.update(params)
        return str(success_url)
